import abc
import time
from typing import Optional, Any

import torch.nn as nn
import torch.optim
from lightning import Fabric
from torch.nn.modules.loss import _Loss

from ncalib import NCA
from ncalib.nca.modular_nca import ModularNCA
from ncalib.trainer.logger import NCA_Logger
from ncalib.trainer.stopping_criterion import NCA_StoppingCriterion
from ncalib.trainer.stopping_criterion.simple_criteria import MaxEpochs
from ncalib.trainer.target import NCA_Target
from ncalib.utils.utils import NCA_Base


class NCA_Trainer(NCA_Base, abc.ABC):
    def __init__(
            self,
            nca: ModularNCA,
            *,
            optimizer: Optional[torch.optim.Optimizer] = None,  # Defaults to Adam with lr=1e-3
            scheduler: Optional[Any] = None,  # Needs to be of type torch.optim._LRScheduler
    ):
        super().__init__()
        self.nca = nca

        # PyTorch training
        if optimizer is None:
            optimizer = torch.optim.Adam(self.nca.parameters(), lr=1e-3)

        self.optimizer = optimizer
        self.scheduler = scheduler

        # Optimize with Fabric
        torch.set_float32_matmul_precision("high")
        self.fabric = Fabric(precision="16-mixed")  # bf16-mixed for 4090, otherwise 16-mixed
        self.nca, self.optimizer = self.fabric.setup(self.nca, self.optimizer)

        # Used during training
        self.losses: list[float] = []
        self._epoch: Optional[int] = None
        self._time_started: Optional[float] = None
        self._time_finished: Optional[float] = None
        self._latest_metrics: dict[str, float] = {}

    @property
    def latest_metrics(self) -> dict[str, float]:
        metrics = {
            "loss": self.latest_loss
        }
        metrics.update(self._latest_metrics)
        return metrics

    @property
    def epoch(self) -> int:
        if self._epoch is None:
            raise RuntimeError("Training not started yet")
        return self._epoch

    @property
    def latest_loss(self) -> float:
        if len(self.losses) == 0:
            raise RuntimeError("Training not started yet")
        return self.losses[-1]

    @property
    def training_duration(self) -> float:
        if self._time_started is None:
            raise RuntimeError("Training not started yet")
        if self._time_finished is None:
            return time.perf_counter() - self._time_started
        return self._time_finished - self._time_started

    @abc.abstractmethod
    def train(
            self,
            stopping_criterion: NCA_StoppingCriterion = MaxEpochs(1000),
            *,
            loggers: NCA_Logger | None = None
    ):
        pass

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["nca"] = self.nca.to_config_dict()
        cfg["normalize_gradients"] = self.normalize_gradients
        return cfg



from functools import wraps
from typing import Sequence

import torch


def discretize(bins: Sequence[float] = (1.5 / 255, 2.5 / 255, 3.5 / 255)):
    this_bins = bins

    @wraps(discretize)
    def wrapped(image: torch.Tensor):
        bins = torch.Tensor(this_bins)
        binned_labels = torch.bucketize(image, bins).squeeze()
        return binned_labels

    return wrapped



import abc

import torch

from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.utils import NCA_Base


class NCA_Target(NCA_Base, abc.ABC):
    def __init__(
            self,
            *,
            device: torch.device | str = DEFAULT_DEVICE
    ):
        super().__init__()

        self.device = device

    @abc.abstractmethod
    def __call__(self, n=1) -> torch.Tensor:
        """ Samples next targets [n x C x W x H]"""
        pass

    @property
    def channels(self) -> int:
        return self().shape[1]

    @property
    def width(self) -> int:
        return self().shape[2]

    @property
    def height(self) -> int:
        return self().shape[3]


class DummyTarget(NCA_Target):
    def __init__(self, channels=4, width=32, height=32, *, device: torch.device | str = DEFAULT_DEVICE):
        super().__init__(device=device)
        self._channels = channels
        self._width = width
        self._height = height

    def __call__(self, n=1) -> torch.Tensor:
        return torch.zeros((n, self._channels, self._width, self._height), device=self.device)



import abc
from typing import Type, Any, Iterable

import torch
import torch.nn as nn
from hydra.utils import instantiate

from ncalib.utils.utils import NCA_Base


class NCA_Perception(NCA_Base, nn.Module, abc.ABC):
    def __init__(self, channel_slice: slice = slice(None, None, None)):
        super().__init__()
        nn.Module.__init__(self)
        self.channel_slice = channel_slice

    def __add__(self, other: "NCA_Perception") -> "ChainedPerception":
        return ChainedPerception(perceptions=[self, other])

    def __getitem__(self, item: slice | int):
        if isinstance(item, int):
            item = slice(item, item + 1)

        if not (self.channel_slice is None or self.channel_slice == slice(None, None, None)) and item != self.channel_slice:
            raise ValueError("This perception already has a channel slice!")

        assert isinstance(item, slice)
        cpy = instantiate(self.to_config_dict())
        cpy.channel_slice = item
        return cpy

    def get_state(self, state: torch.Tensor) -> torch.Tensor:
        return state[:, self.channel_slice]

    @abc.abstractmethod
    def forward(self, state: torch.Tensor, **kwargs) -> list[torch.Tensor]:
        """
        :param state: Tensor [B x C x W x H]
        :return: List of Tensor [B x C x W x H]
        """
        pass

    def output_size_for_n_channels(self, channels: int) -> int:
        img = torch.zeros((1, channels, 1, 1))
        perception = self(img)
        return len(self.concat(perception)[0])

    @classmethod
    def concat(cls, perceptions: list[torch.Tensor]) -> torch.Tensor:
        """
        Concatenates a list of perceptions as one tensor in the channel dimension
        :param perceptions: list of Tensor [B x C x W x H]
        :return: [B x C*n x W x H]
        """
        return torch.cat(perceptions, dim=1)

    def required_kwargs(self) -> set[str]:
        return set()

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["channel_slice"] = self.channel_slice
        return cfg


class ChainedPerception(NCA_Perception):
    def __init__(
            self,
            *,
            perceptions: Iterable[Type[NCA_Perception] | NCA_Perception],
            channel_slice: slice = slice(None, None, None)
    ):
        super().__init__(channel_slice=channel_slice)
        self.perceptions: tuple[perceptions] = self._unchain(tuple(perceptions))

    def forward(self, state: torch.Tensor, **kwargs) -> list[torch.Tensor]:
        perceptions = []
        state = self.get_state(state)
        for perception in self.perceptions:
            perceptions += perception(state, **kwargs)
        return perceptions

    def _unchain(
            self,
            perceptions: tuple[NCA_Perception | Type[NCA_Perception]],
    ) -> tuple[NCA_Perception]:
        """
        Unchains perceptions, so that chained perceptions do not have chained perceptions as perceptions.
        E.g.
        Input: (SobelXPerception(), ChainedPerception(perceptions=[SobelYPerception(), LaPlacePerception()]))
        Output: (SobelXPerception(), SobelYPerception(), LaPlacePerception())

        Input: (SobelXPerception(), SobelYPerception(), LaPlacePerception())
        Output: (SobelXPerception(), SobelYPerception(), LaPlacePerception())
        """
        res = []
        for perception in perceptions:
            if isinstance(perception, type):
                if not issubclass(perception, NCA_Perception):
                    raise TypeError(f"Perception has to be a subclass of {NCA_Perception} (Got {perception})")
                res.append(perception())
            if isinstance(perception, ChainedPerception):
                res += list(self._unchain(perception.perceptions))
            elif isinstance(perception, NCA_Perception):
                res.append(perception)
            else:
                TypeError(f"Unknown type for {perception}: {type(perception)}")
        return tuple(res)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["perceptions"] = [perception.to_config_dict() for perception in self.perceptions]
        return cfg

    def required_kwargs(self) -> set[str]:
        res = set()
        for perception in self.perceptions:
            res.update(perception.required_kwargs())

        return res

    def reset(self):
        for perception in self.perceptions:
            perception.reset()



import abc
import datetime
import logging
import sys
import warnings
from pathlib import Path
from typing import Optional, Any

import torch
import torch.nn as nn
from hydra.utils import instantiate
from importlib_metadata import version, PackageNotFoundError

from ncalib.nca.perception import NCA_Perception, ChainedPerception
from ncalib.utils.torch_helper import cleanup_state_dict
from ncalib.utils.utils import full_class_name, get_git_state, NCA_Base


class NCA(nn.Module, NCA_Base, abc.ABC):
    def __init__(
            self,
            *,
            device: torch.device | str = torch.device('cpu')
    ):
        super().__init__()
        NCA_Base.__init__(self)
        self._last_delta_state: Optional[torch.Tensor] = None

        self.device = torch.device(device)

    @property
    @abc.abstractmethod
    def channels(self) -> int:
        pass

    def reset(self):
        raise NotImplementedError(f"reset missing in {self.__class__.__name__}")

    @property
    def last_delta_state(self) -> torch.Tensor:
        """
        :return: torch.Tensor[B x C x W x H]
        """
        if self._last_delta_state is None:
            raise RuntimeError("No delta state calculated yet")
        return self._last_delta_state

    def required_kwargs(self) -> set[str]:
        return set()

    def save_as(self, filepath: Path | str) -> dict[str, Any]:
        cfg = self.as_full_config_dict()
        torch.save(cfg, filepath)
        return cfg

    def as_full_config_dict(self) -> dict[str, Any]:
        """
        Returns a complete config dict including the architecture, the state dict and some additional metadata
        """
        state_dict = self.state_dict()
        architecture = self.to_config_dict()
        cfg = {
            "architecture": architecture,
            "state_dict": state_dict,
            "required_kwargs": self.required_kwargs(),
            "meta": self.generate_meta_data()
        }
        return cfg

    @classmethod
    def load_from_path(cls, filepath: Path | str) -> "NCA":
        logging.getLogger("NCA_LOAD").info(f"Loading from path {Path(filepath).absolute()}")
        full_config_dict = torch.load(filepath, weights_only=False)
        return cls.load_from(full_config_dict)

    @classmethod
    def load_from_wandb(cls, entity: str, project: str, run_id: str, model_epoch: Optional[Any] = None) -> "NCA":
        import wandb
        from ncalib.utils.wandb import get_wandb_model

        api = wandb.Api()
        run = api.run(f"{entity}/{project}/{run_id}")
        assert isinstance(run, wandb.apis.public.Run)
        config = run.config
        config["state_dict"] = get_wandb_model(run, model_epoch=model_epoch)

        return cls.load_from(config)

    @classmethod
    def load_from(cls, full_config_dict: dict[str, Any]) -> "NCA":
        logger = logging.getLogger("NCA_LOAD")
        # Access Meta
        meta = full_config_dict["meta"]

        # Check versions
        for package, _version in meta["packages"].items():
            try:
                current_version = version(package)
            except PackageNotFoundError:
                current_version = None
            if current_version != _version:
                warnings.warn(
                    f'Package version mismatch! '
                    f'"{package}" has version "{current_version}", but you loaded model used version "{_version}". '
                    f'This might not work!'
                )

        logger.info(
            f"Loading NCA from {meta['date']}. Required kwargs: {list(full_config_dict.get('required_kwargs', []))}")

        # Load NCA
        if "trainer" in full_config_dict:
            architecture = full_config_dict["trainer"]["nca"]
        elif "architecture" in full_config_dict:
            architecture = full_config_dict["architecture"]
        else:
            raise ValueError("Could not find architecture in config dict")

        nca = instantiate(architecture)
        if "state_dict" in full_config_dict and full_config_dict["state_dict"] is not None:
            state_dict = cleanup_state_dict(full_config_dict["state_dict"])
            nca.load_state_dict(state_dict)

        return nca

    def generate_meta_data(self) -> dict[str, Any]:
        package_versions = {}
        for package in ["ncalib", "torch", "numpy"]:
            try:
                v = version(package)
            except:
                v = "UNKNOWN"
            package_versions[package] = v

        return {
            "date": datetime.datetime.now(),
            "git": get_git_state(),
            "python_version": sys.version,
            "num_params": self.num_params,
            "packages": package_versions
        }

    @property
    def num_params(self):
        return sum(p.numel() for p in self.parameters())



import logging
import math
import random
import warnings
from dataclasses import is_dataclass, asdict
from pathlib import Path
from typing import Any, Literal, Callable, Optional, Iterable, Iterator

import git
import numpy as np
import torch
import torchvision
from flatten_dict import flatten
from git import InvalidGitRepositoryError
from omegaconf import OmegaConf, DictConfig
from torch.backends import cudnn
from torch.utils.data import Dataset

from ncalib.utils.torch_helper import SimpleImageDataset


class NCA_Base:
    LOG_LEVEL = logging.DEBUG

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.level = NCA_Base.LOG_LEVEL

    def __repr__(self):
        init_code = self.__init__.__code__
        args = ", ".join([
            f"{k}={getattr(self, k, '???')}"
            for k in init_code.co_varnames[1:init_code.co_argcount]
        ])
        return f"{self.__class__.__name__}({args})"

    def reset(self):
        pass

    def to_config_dict(self) -> dict[str, Any]:
        return {"_target_": full_class_name(self)}


def add_prefix_to_keys(obj: dict[str, Any], prefix: str) -> dict[str, Any]:
    return {prefix + key: value for key, value in obj.items()}


def config_as_flat_dict(
        cfg: Any,
        *,
        reducer: Literal["underscore", "tuple", "dot", "path"] | Callable[[Any, Any], str] = "underscore"
) -> dict[str, Any]:
    # Make dict
    if is_dataclass(cfg):
        cfg_dict = asdict(cfg)
    elif isinstance(cfg, (OmegaConf, DictConfig)):
        cfg_dict = OmegaConf.to_container(cfg, resolve=True, enum_to_str=True)
    else:
        cfg_dict = dict(cfg)

    # Flatten
    return flatten(cfg_dict, reducer=reducer)


def own_logspace(start: float, end: float, steps: int) -> torch.Tensor:
    return torch.logspace(math.log10(start), math.log10(end), steps)


def fix_seed(seed: Optional[int], *, deterministic=True):
    logger = logging.getLogger("Seed")
    if seed is None:
        logger.warning(f"No seed provided!")
        return

    logger.info(f"Fix seed to {seed}.")
    torch.manual_seed(seed)
    np.random.seed(seed + 1)
    random.seed(seed + 2)

    if deterministic:
        cudnn.deterministic = True
        logger.warning("You have chosen to seed training. "
                       "This will turn on the CUDNN deterministic setting, "
                       "which can slow down your training considerably! "
                       "You may see unexpected behavior when restarting "
                       "from checkpoints.")


def endless_generator(iterable: Iterable) -> Iterator:
    while True:
        for element in iterable:
            yield element


def length_of_slice(s: slice) -> int:
    if not s.stop:
        raise ValueError(f"Cannot get length of indefinite slice: {s}")
    n = s.stop
    if s.start:
        n -= s.start
    if s.step:
        n = int(n // s.step)

    return n


def full_class_name(o: object) -> str:
    # https://stackoverflow.com/a/13653312
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:  # Check for __builtin__
        return o.__class__.__name__
    return f'{module}.{o.__class__.__name__}'


def slice_to_config(s: slice) -> Optional[dict[str, Any]]:
    if s is None:
        return None
    return {
        "_target_": f"{to_slice.__module__}.{to_slice.__name__}",
        "start": s.start,
        "stop": s.stop,
        "step": s.step,
    }


def optimizer_to_config(optimizer: torch.optim.Optimizer) -> dict[str, Any]:
    cfg = {
        "_target_": full_class_name(optimizer),

    }
    cfg.update(optimizer.defaults)
    return cfg


def scheduler_to_config(scheduler: torch.optim.lr_scheduler.LRScheduler) -> Optional[dict[str, Any]]:
    if scheduler is None:
        return None

    cfg = {
        "_target_": full_class_name(scheduler),
    }
    for var in vars(scheduler.__class__):  # Needs to get var names from scheduler.__class__, but values from scheduler
        cfg[var] = getattr(scheduler, var)
    return cfg


def dataset_to_config(dataset: Dataset) -> dict[str, Any]:
    cfg = {
        "_target_": full_class_name(dataset)
    }
    if hasattr(dataset, "transforms"):
        cfg["transforms"] = transforms_to_config(dataset.transforms)

    if isinstance(dataset, SimpleImageDataset):
        cfg["paths"] = dataset.paths

    return cfg


def transforms_to_config(transform) -> dict[str, Any]:
    cfg = {
        "_target_": full_class_name(transform)
    }
    if isinstance(transform, (torchvision.transforms.Compose,)):
        cfg["transforms"] = [transforms_to_config(t) for t in transform.transforms]
    if isinstance(transform, (torch.nn.ModuleList, torch.nn.Sequential)):
        cfg["transforms"] = [transforms_to_config(t) for t in transform.modules()]

    return cfg


def to_slice(start, stop, step):
    return slice(start, stop, step)


def get_git_state(path: Optional[Path | str] = None) -> dict[str, str]:
    if path is None:
        path = __file__
    path = Path(path)
    repo = None
    while path.parent != path:
        try:
            repo = git.Repo(path)
            break
        except InvalidGitRepositoryError:
            path = path.parent

    if repo is None:
        warnings.warn("No Git Repository found!")
        return {}

    name = Path(repo.working_dir).name
    commit = repo.head.commit

    return {
        "repository": name,
        "commit": commit.hexsha,
        "is_dirty": repo.is_dirty()

    }


def add_to_dict_key(data: dict[str, Any], *, prefix: str = "", suffix: str = "") -> dict[str, Any]:
    """Adds a prefix and a suffix to every key in a dictionary"""
    return {f"{prefix}{key}{suffix}": value for key, value in data.items()}


def to_binary_list(n: int, *, min_length=0) -> list[int]:
    """ Converts an integer into its binary representation as a list of 1 or 0 (MSB first) """
    result = list(map(int, list(f"{n:b}")))
    return [0] * max(min_length - len(result), 0) + result


def make_circle_masks(b, w, h) -> torch.Tensor:
    """
    Adapted from https://github.com/jstovold/ALIFE2023/blob/master/InternalSignals.ipynb

    :returns: Tensor[B, W, H]
    """
    # TODO: RNG is not seeded in this function.
    # TODO: Check W,H in correct order. Doesnt matter with square shapes
    x = torch.linspace(-1.0, 1.0, w).unsqueeze(0).unsqueeze(0)
    y = torch.linspace(-1.0, 1.0, h).unsqueeze(1).unsqueeze(0)
    center = torch.rand(2, b, 1, 1) * 0.9 + 0.05  # random between 0.05,...,0.95
    r = torch.rand(b, 1, 1) * 0.3 + 0.1
    x, y = (x - center[0]) / r, (y - center[1]) / r
    mask = (x * x + y * y < 1.0).float()
    return mask



from typing import Optional, Any

import torch
from torch.nn import functional as F
from torch.utils.data import Dataset, DataLoader

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.trainer.dataset import NCA_SteppedDataset, DataSample
from ncalib.utils.torch_helper import EndlessIterableDataset


class ClassificationDataset(NCA_SteppedDataset):
    def __init__(
            self,
            dataset: Dataset,  # Expect to yield tuple[image, label], where label is number of class
            num_classes: int,
            seed_factory: NCA_SeedFactory,
            *,
            step_range: tuple[int, int],
            use_one_hot_encoding: bool = False,
            dataset_channel_offset: int = 0,
            seed: Optional[int] = None,
            data_sample_base_class = DataSample,
    ):
        super().__init__(step_range, seed=seed)
        self.dataset = dataset
        self.dataset_dataloader = iter(
            DataLoader(
                EndlessIterableDataset(dataset, generator=self.rng, shuffle=True),
            )
        )
        self.dataset_channel_offset = dataset_channel_offset
        self.num_classes = num_classes
        self.seed_factory = seed_factory
        self.use_one_hot_encoding = use_one_hot_encoding
        self.data_sample_base_class = data_sample_base_class

    def __call__(self) -> DataSample:
        num_step = self.generate_n_steps()

        image, original_label = next(self.dataset_dataloader)
        B, C, W, H = image.shape
        # image: [B=1 x C x W x H]; label: [B] or [B x W x H]
        state = self.seed_factory(1, device="cpu")
        state[:, self.dataset_channel_offset:self.dataset_channel_offset + C] = image

        if len(original_label.shape) == 1:  # Only a single class per batch -> Reshape to image dimensions
            label = original_label[:, None, None].repeat(B, W, H)
        elif len(original_label.shape) == 4 and original_label.shape[1] == 1:  # Label has a squeezable dimension
            label = original_label[:, 0]
        else:
            label = original_label

        if not ((B, W, H) == label.shape):
            raise ValueError(
                f"Invalid shapes. "
                f"Image shape: {image.shape}, "
                f"Label shape: {label.shape} "
                f"(original: {original_label.shape})"
            )

        if B != 1:
            raise RuntimeError(f"Batch size should be 1. (is {B}; Shape: {image.shape})")

        # label: [B x W x H]
        if self.use_one_hot_encoding:
            target = F.one_hot(label.to(torch.int64), num_classes=self.num_classes)
            # target: Tensor[1 x W x H x N]
            target = target.permute(0, 3, 1, 2)
            # target: Tensor[1 x N x W x H]
        else:
            target = label

        return self.data_sample_base_class(
            state,
            torch.as_tensor(target, dtype=torch.float),
            num_step,
            batch_size=[B],
            training_regenerate=torch.as_tensor([False]*B)
        )

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["dataset"] = self.dataset.__class__.__name__
        cfg["num_classes"] = self.num_classes
        cfg["use_one_hot_encoding"] = self.use_one_hot_encoding
        cfg["seed_factory"] = self.seed_factory.to_config_dict()
        cfg["dataset_channel_offset"] = self.dataset_channel_offset
        return cfg


class GlobalClassificationDataset(NCA_SteppedDataset):
    def __init__(
            self,
            dataset: Dataset,  # Expect to yield tuple[image, label], where label is number of class
            num_classes: int,
            seed_factory: NCA_SeedFactory,
            *,
            step_range: tuple[int, int],
            use_one_hot_encoding: bool = False,

            seed: Optional[int] = None,
    ):
        super().__init__(step_range, seed=seed)
        self.dataset = dataset
        self.dataset_dataloader = iter(
            DataLoader(
                EndlessIterableDataset(dataset, generator=self.rng, shuffle=True),
            )
        )
        self.num_classes = num_classes
        self.seed_factory = seed_factory
        self.use_one_hot_encoding = use_one_hot_encoding

    def __call__(self) -> DataSample:
        num_step = self.generate_n_steps()

        image, original_label = next(self.dataset_dataloader)
        B, C, W, H = image.shape
        # image: [B=1 x C x W x H]; label: [B] or [B x W x H]
        state = self.seed_factory(1, device="cpu")
        state[:, :C] = image

        if len(original_label.shape) == 1:  # Only a single class per batch -> Reshape to image dimensions
            label = original_label[:, None, None].repeat(B, W, H)
        elif len(original_label.shape) == 4 and original_label.shape[1] == 1:  # Label has a squeezable dimension
            label = original_label[:, 0]
        else:
            label = original_label

        if not ((B, W, H) == label.shape):
            raise ValueError(
                f"Invalid shapes. "
                f"Image shape: {image.shape}, "
                f"Label shape: {label.shape} "
                f"(original: {original_label.shape})"
            )

        if B != 1:
            raise RuntimeError(f"Batch size should be 1. (is {B}; Shape: {image.shape})")

        # label: [B x W x H]
        if self.use_one_hot_encoding:
            target = F.one_hot(label.to(torch.int64), num_classes=self.num_classes)
            # target: Tensor[1 x W x H x N]
            target = target.permute(0, 3, 1, 2)
            # target: Tensor[1 x N x W x H]
        else:
            target = label

        return DataSample(
            state,
            torch.as_tensor(target, dtype=torch.float),
            num_step,
            batch_size=[B],
            training_regenerate=False
        )



import abc

import torch
from torch import nn

from ncalib.utils.utils import NCA_Base


class NCA_StateUpdater(NCA_Base, nn.Module, abc.ABC):
    """
    Takes a state and a state update as input and returns the next state. E.g. normal Update and Alive-Masking Update
    """

    def __init__(self):
        super().__init__()
        nn.Module.__init__(self)
        self.model_size_factor = 1

    @abc.abstractmethod
    def forward(
            self,
            state: torch.Tensor,
            delta_state: torch.Tensor,
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        :return: updated state, updated delta_state
        """
        pass

    def __add__(self, other: "NCA_StateUpdater"):
        # Prepare each state updater so that each is an iterable of simple state updaters
        if isinstance(self, CombinedStateUpdater):
            states1 = self.state_updater
        else:
            states1 = [self]

        if isinstance(other, CombinedStateUpdater):
            states2 = other.state_updater
        else:
            states2 = [other]

        # Combine State Updaters
        return CombinedStateUpdater(*states1, *states2)

    def required_kwargs(self) -> set[str]:
        return set()


class CombinedStateUpdater(NCA_StateUpdater):
    def __init__(self, *state_updater: NCA_StateUpdater):
        super().__init__()
        self.state_updater = nn.ModuleList(state_updater)
        self.model_size_factor = 1
        for updater in state_updater:
            self.model_size_factor *= updater.model_size_factor

    def forward(
            self,
            state: torch.Tensor,
            delta_state: torch.Tensor,
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        for state_updater in self.state_updater:
            state, delta_state = state_updater(state, delta_state, **kwargs)

        return state, delta_state

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["_args_"] = [
            state_updater.to_config_dict()
            for state_updater in self.state_updater
        ]
        return cfg

    def reset(self):
        for state_updater in self.state_updater:
            state_updater.reset()

    def required_kwargs(self) -> set[str]:
        res = super().required_kwargs()
        for state_updater in self.state_updater:
            res.update(state_updater.required_kwargs())
        return res



import abc
from typing import Optional

import torch

from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.utils import NCA_Base


class NCA_SeedFactory(NCA_Base, abc.ABC):
    def __init__(
            self,
            width: int = 64,
            channels: int = 16,
            height: Optional[int] = None,
    ):
        super().__init__()
        self.width = width
        self.height = height if height is not None else width
        self.channels = channels

    @abc.abstractmethod
    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE) -> torch.Tensor:
        pass

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["width"] = self.width
        cfg["height"] = self.height
        cfg["channels"] = self.channels
        return cfg


class NCA_SeedPostprocessing(NCA_SeedFactory, abc.ABC):
    def __init__(self, seed_factory: NCA_SeedFactory):
        super().__init__()
        self.base_seed_factory = seed_factory

    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE):
        seed = self.base_seed_factory(n, device=device)
        return self.update_seed(seed)

    @abc.abstractmethod
    def update_seed(self, original_seed: torch.Tensor) -> torch.Tensor:
        pass

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["seed_factory"] = self.base_seed_factory
        return cfg



from typing import Any
from ncalib.utils.torch_helper import weight_init

import torch
from torch import nn

from ncalib.nca.models import NCA_Model


class LinearNCAModel(NCA_Model):
    def __init__(self, input_size: int, output_size: int, hidden_channels: int = 128, hidden_layers: int = 1, out_bias=True):
        super().__init__(input_size, output_size)
        layers = []
        for i in range(hidden_layers):
            layers += [
                nn.Conv2d(input_size, hidden_channels, kernel_size=1),
                nn.ReLU(),
            ]
            input_size = hidden_channels

        self.conv_hidden = nn.Sequential(*layers)

        # Final output layer
        self.out_bias = out_bias
        self.conv_output = nn.Conv2d(input_size, output_size, kernel_size=1, bias=self.out_bias)

        self.hidden_channels = hidden_channels
        self.hidden_layers = hidden_layers

    def forward(self, perception: list[torch.Tensor], **kwargs) -> torch.Tensor:
        total_perception = torch.cat(perception, 1)
        delta_state = self.conv_hidden(total_perception)
        delta_state = self.conv_output(delta_state)

        return delta_state

    def reset(self):
        self.apply(weight_init)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["hidden_channels"] = self.hidden_channels
        cfg["hidden_layers"] = self.hidden_layers
        cfg["out_bias"] = self.out_bias
        return cfg



from typing import Any, Optional

import torch
import torch.nn.functional as F
from jaxtyping import Float

from ncalib.nca.state_updater import NCA_StateUpdater
from ncalib.utils.utils import slice_to_config


class AliveMasking(NCA_StateUpdater):
    def __init__(
            self,
            *,
            threshold: float = 0.1,
            max_pool_size: int = 3,
            alpha_channel: int = 3,
            ignore_channels: Optional[slice] = None
    ):
        super().__init__()
        self.threshold = threshold
        self.max_pool_size = max_pool_size

        self.alpha_channel = alpha_channel
        self.ignore_channels= ignore_channels

    def forward(
            self,
            state: Float[torch.Tensor, "B C W H"],
            delta_state: Float[torch.Tensor, "B C W H"],
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        :param state: Current State of the NCA
        :param delta_state: Update for NCA
        :return: Next state of NCA
        """
        # Get Alive Mask
        max_pool = F.max_pool2d(state[:, self.alpha_channel, :, :], self.max_pool_size, stride=1, padding=self.max_pool_size // 2)
        alive = max_pool > self.threshold
        alive = torch.unsqueeze(alive, 1)
        alive = torch.repeat_interleave(alive, state.shape[1], 1)

        if self.ignore_channels is not None:
            alive[:, self.ignore_channels] = 1

        # Apply Alive Mask
        next_state = state * alive
        return next_state, delta_state

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["threshold"] = self.threshold
        cfg["max_pool_size"] = self.max_pool_size
        cfg["alpha_channel"] = self.alpha_channel
        cfg["ignore_channels"] = slice_to_config(self.ignore_channels)

        return cfg



from typing import Literal

import numpy as np
import torch
import torch.nn.functional as F


def simple_2d_convolution(image, kernel):
    """
    :param image:  Tensor[W x H]
    :param kernel: Tensor[W' x H']
    :return: Tensor[W - W'//2 x H - H'//2]
    """
    return F.conv2d(
        torch.unsqueeze(torch.unsqueeze(image, dim=0), dim=0).float(),
        kernel.repeat(1, 1, 1, 1).float(), groups=1, padding=0
    )[0, 0]


def batched_convolution_2D(
        source: torch.Tensor,
        kernel: torch.Tensor,
        padding_mode: Literal["constant", "reflect", "replicate", "circular"] = "replicate"
) -> torch.Tensor:
    """
    :param source: Tensor[B x N x W x H]
    :param kernel: Tensor[W' x H']
    :param padding_mode:
    :return: Tensor[B x N x W x H]
    """
    # Check input sizes
    if not len(kernel.shape) == 2:
        raise ValueError("Kernel has to be 2D (W x H)")

    if not len(source.shape) == 4:
        raise ValueError("Source has to be 4D (B x n x W x H)")

    n_metrics = source.shape[1]
    kernel_width, kernel_height = kernel.shape

    # Calculate padding
    top_pad = kernel_height // 2
    left_pad = kernel_width // 2
    bottom_pad = kernel_height - top_pad - 1
    right_pad = kernel_width - left_pad - 1

    # Convolution
    kernel = kernel.repeat(n_metrics, 1, 1, 1).float().to(source.device)
    padded = F.pad(source.float(), (top_pad, bottom_pad, left_pad, right_pad), mode=padding_mode)
    result = F.conv2d(padded, kernel, groups=n_metrics, padding=0)

    # Validate result
    if result.shape != source.shape:
        raise ArithmeticError(f"Something went wrong. Shape mismatch: src={source.shape}, res={result.shape}")

    return result


with torch.no_grad():
    # LaPlace
    # https://jblindsay.github.io/ghrg/Whitebox/Help/FilterLaplacian.html
    laplace_3x3_1 = torch.Tensor([[-1., -1., -1.], [-1., 8., -1.], [-1., -1., -1.]])
    laplace_3x3_2 = torch.Tensor([[0., -1., 0], [-1., 4., -1.], [0, -1., 0]])

    laplace_5x5 = torch.Tensor([
        [0., 0., -1., 0., 0.],
        [0., -1., -2., -1., 0.],
        [-1., -2., 17., -2., -1.],
        [0., -1., -2., -1., 0.],
        [0., 0., -1., 0., 0.],
    ])

    # Sobol
    sobol_x = torch.tensor([[1., 0, -1.], [2., 0, -2.], [1., 0, -1.]])
    sobol_y = torch.tensor([[1., 2., 1.], [0., 0., 0.], [-1., -2., -1.]])
    sobol_x5 = torch.tensor([
        [-1, -2, 0, 2, 1],
        [-4, -8, 0, 8, 4],
        [-6, -12, 0, 12, 6],
        [-4, -8, 0, 8, 4],
        [-1, -2, 0, 2, 1]
    ])
    sobol_y5 = torch.tensor([
        [-1, -4, -6, -4, -1],
        [-2, -8, -12, -8, -2],
        [0, 0, 0, 0, 0],
        [2, 8, 12, 8, 2],
        [1, 4, 6, 4, 1]
    ])

    zeros_3x3 = torch.zeros((3, 3))


# Gaussian Blur
def gaussian_blur(filter_size: int = 3, std: float = 3) -> torch.Tensor:
    if not (filter_size % 2) == 1:
        raise ValueError("Gaussian blur only works for odd filter sizes! (at least in this implementation...)")

    center = (filter_size - 1) / 2
    sigma2 = std ** 2
    i, j = np.indices((filter_size, filter_size), dtype=np.float64) - center

    filter_ = torch.as_tensor(np.exp(-(i ** 2 + j ** 2) / (2 * sigma2)) / (2 * np.pi * sigma2))
    return filter_.float() / torch.sum(filter_)


def block_kernel(size: int = 3) -> torch.Tensor:
    return torch.ones((size, size)).float()


# Surrounding mean
surrounding_mean_3x3 = torch.tensor([[2., 1., 2.], [1., 0., 1.], [2., 1., 2.]])
surrounding_mean_5x5 = torch.Tensor([
    [8., 4., 2., 4., 8.],
    [4., 2., 1., 2., 4.],
    [2., 1., 0., 1., 2.],
    [4., 2., 1., 2., 4.],
    [8., 4., 2., 4., 8.],
])

inward_mean_3x3 = torch.max(surrounding_mean_3x3) - surrounding_mean_3x3
inward_mean_5x5 = torch.max(surrounding_mean_5x5) - surrounding_mean_5x5

surrounding_mean_3x3 = surrounding_mean_3x3 / torch.sum(surrounding_mean_3x3)
surrounding_mean_5x5 = surrounding_mean_5x5 / torch.sum(surrounding_mean_5x5)
inward_mean_3x3 = inward_mean_3x3 / torch.sum(inward_mean_3x3)
inward_mean_5x5 = inward_mean_5x5 / torch.sum(inward_mean_5x5)



from typing import Any

import torch

from ncalib.nca.state_updater import NCA_StateUpdater


class AsynchronousUpdate(NCA_StateUpdater):
    def __init__(self, *, update_probability: float = 0.5):
        super().__init__()
        self.update_probability = update_probability

    def forward(
            self,
            state: torch.Tensor,
            delta_state: torch.Tensor,
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        n, c, w, h = delta_state.shape

        # Create random Mask
        random_mask = torch.rand((n, 1, w, h), device=delta_state.device)
        random_mask = random_mask < self.update_probability

        # State Update
        next_delta = delta_state * random_mask
        return state, next_delta

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["update_probability"] = self.update_probability
        return cfg



from typing import Optional

import torch

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.utils import slice_to_config


class IsotropicSeedFactory(NCA_SeedFactory):
    def __init__(
            self,
            width: int = 64,
            channels: int = 16,
            *,
            height: Optional[int] = None,
            angle_channel_id: int = 0,
            ignore_channels: Optional[slice] = None,
        limit_possible_values: Optional[int] = None,
    ):
        super().__init__(width=width, height=height, channels=channels)
        self.ignore_channels = ignore_channels
        self.angle_channel_id = angle_channel_id
        self.limit_possible_values = limit_possible_values

    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE) -> torch.Tensor:
        # Init with zeros
        seed = torch.zeros([n, self.channels, self.width, self.height], dtype=torch.float32, device=device)

        # Set center to 1
        mask_for_ones = torch.ones(self.channels, dtype=torch.bool)
        if self.ignore_channels is not None:
            mask_for_ones[self.ignore_channels] = False
        mask_for_ones[self.angle_channel_id] = False
        seed[:, mask_for_ones, self.width // 2, self.height // 2] = 1.0

        # Initialize random angles
        random_values = torch.rand((n, self.width, self.height))
        if self.limit_possible_values is not None:
            random_values = (random_values // (1 / self.limit_possible_values)) / self.limit_possible_values
        seed[:, self.angle_channel_id] = random_values * torch.pi * 2.0
        return seed

    @classmethod
    def for_target(cls, target: "NCA_Target", nca: "NCA", *, angle_channel_id: int = 0) -> "IsotropicSeedFactory":
        return cls(
            width=target.width,
            height=target.height,
            channels=nca.channels,
            angle_channel_id=angle_channel_id
        )

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["ignore_channels"] = slice_to_config(self.ignore_channels)
        cfg["angle_channel_id"] = self.angle_channel_id
        cfg["limit_possible_values"] = self.limit_possible_values
        return cfg



import cv2
import torch

from ncalib.utils.utils import length_of_slice
from ncalib.visualization.state_visualizer import BatchedStateVisualizer
import colorsys

def get_colormap(n, device="cpu"):
    # Evenly space n hues over [0,1), full saturation and brightness.
    colormap = [colorsys.hsv_to_rgb(i / n, 1.0, 1.0) for i in range(n)]
    # Return tensor with shape [3, n] so that indexing with labels gives [3, W, H]
    return torch.tensor(colormap, device=device, dtype=torch.float32)

class ClassificationVisualizer(BatchedStateVisualizer):
    def __init__(
            self,
            channels: slice,
            *,
            colormap: int = cv2.COLORMAP_SUMMER,
            name: str = "Classification",
            batch_slice=slice(None, None, None),
            mask_channels: slice | int = 0
    ):
        super().__init__(name=name, batch_slice=batch_slice)

        self.channels = channels
        if isinstance(mask_channels, int):
            mask_channels = slice(mask_channels, mask_channels + 1)
        self.mask_channels = mask_channels
        self.num_classes = length_of_slice(self.channels)

        self.colormap = get_colormap(self.num_classes).T

    def _batch_to_visualization(self, batch: torch.Tensor) -> dict[str, torch.Tensor]:
        """
        :param batch: Tensor[B x C x W x H]
        :return: dict[name -> Tensor[B x 1 x W x H] with float ranges 0..1 in Gray
        """
        batched_results = []
        batched_channels = batch[:, self.channels]
        masked_channels = batch[:, self.mask_channels]

        B, C, W, H = batch.shape

        self.colormap = self.colormap.to(batch.device)

        for channels, mask in zip(batched_channels, masked_channels):
            # softmax = torch.softmax(channels, dim=0)
            arg_max = torch.argmax(channels, dim=0)
            result = torch.zeros((3, W, H), device=batch.device)
            result += self.colormap[:, arg_max]
            # result = torch.einsum("sij, sc-> cij", softmax, colors)
            result *= torch.mean(mask, dim=0)
            batched_results.append(result)

        return {self.name: torch.stack(batched_results, dim=0)}



import abc
from typing import Literal, Any

import torch
import torch.nn as nn
import torch.nn.functional as F

from ncalib.nca.perception import NCA_Perception
from ncalib.utils import kernels
from ncalib.utils.torch_helper import weight_init


class IdentityPerception(NCA_Perception):
    def forward(self, state: torch.Tensor, **kwargs) -> list[torch.Tensor]:
        return [self.get_state(state)]

    def to(self, device: torch.device | str) -> "IdentityPerception":
        return self


class FilterPerception(NCA_Perception, abc.ABC):
    KERNEL: torch.Tensor

    def __init__(
            self,
            *,
            padding_mode: Literal["constant", "reflect", "replicate", "circular"] = "replicate",
            channel_slice: slice = slice(None, None, None)
    ):
        """
        :param padding_mode: See https://pytorch.org/docs/stable/generated/torch.nn.functional.pad.html for more info
        """
        super().__init__(channel_slice=channel_slice)
        W, H = self.KERNEL.shape
        self.kernel = self.KERNEL.view(1, 1, W, H)
        self.padding_mode = padding_mode
        self._padding = (int(W // 2), int(H // 2)) * 2  # Creates a tuple of 4 values

    def forward(self, state: torch.Tensor, **kwargs) -> list[torch.Tensor]:
        state = self.get_state(state)
        channels = state.shape[1]
        padded = F.pad(state, self._padding, mode=self.padding_mode)
        return [F.conv2d(padded, self.kernel.repeat(channels, 1, 1, 1).to(state.device), groups=channels, padding=0)]

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["padding_mode"] = self.padding_mode
        return cfg


class SobelXPerception(FilterPerception):
    KERNEL = kernels.sobol_x


class SobelYPerception(FilterPerception):
    KERNEL = kernels.sobol_y


class SobelX5Perception(FilterPerception):
    KERNEL = kernels.sobol_x5


class SobelY5Perception(FilterPerception):
    KERNEL = kernels.sobol_y5


class LaPlace1Perception(FilterPerception):
    KERNEL = kernels.laplace_3x3_1


class LaPlace2Perception(FilterPerception):
    """ Version from https://distill.pub/selforg/2021/textures/ """
    KERNEL = kernels.laplace_3x3_2


class ConvPerception(FilterPerception, nn.Module):
    KERNEL = kernels.zeros_3x3

    def __init__(
            self,
            out_channels: int,
            kernel_size: int | tuple[int, int] = (3, 3),
            padding_mode: Literal["constant", "reflect", "replicate", "circular"] = "replicate",
    ):
        super().__init__(padding_mode=padding_mode)
        nn.Module.__init__(self)
        del self.kernel
        self.out_channels = out_channels
        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        self.kernel_size = kernel_size
        self.register_parameter("kernel", nn.Parameter(torch.randn(out_channels, 1, *kernel_size)))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(out_channels={self.out_channels}, kernel_size={self.kernel_size}, padding_mode={self.padding_mode})"

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["out_channels"] = self.out_channels
        cfg["kernel_size"] = self.kernel_size
        return cfg

    def reset(self):
        self.apply(weight_init)



from typing import Any

import torch
import torch.nn as nn
import torch.nn.functional as F

from ncalib import NCA
from ncalib.nca.models.basic import BasicNCAModel
from ncalib.nca.models.tiny import TinyNCAModel
from ncalib.nca.modular_nca import ModularNCA
from ncalib.nca.perception.filter import IdentityPerception, SobelXPerception, SobelYPerception, LaPlace1Perception
from ncalib.nca.state_updater.alive_masking import AliveMasking
from ncalib.nca.state_updater.asynchronous_update import AsynchronousUpdate
from ncalib.nca.state_updater.simple_update import SimpleStateUpdater
from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.kernels import sobol_x, sobol_y


def generate_default_nca(channels=16, *, device=DEFAULT_DEVICE) -> ModularNCA:
    perception = IdentityPerception() + SobelXPerception() + SobelYPerception()

    model = BasicNCAModel(perception.output_size_for_n_channels(channels), channels)

    nca = ModularNCA(
        channels,
        perception,
        model,
        update_function=AsynchronousUpdate() + SimpleStateUpdater() + AliveMasking(),
        device=device
    )
    return nca


def generate_tiny_nca(channels=8, *, device=DEFAULT_DEVICE) -> ModularNCA:
    perception = LaPlace1Perception()

    model = TinyNCAModel(perception.output_size_for_n_channels(channels), channels)

    nca = ModularNCA(channels, perception, model, device=device)
    return nca


class HardcodedNCA(NCA):
    """ Code taken from https://github.com/SethPipho/pytorch-neural-cellular-automata """

    def __init__(
            self,
            channels=16,
            *,
            device: torch.device | str = torch.device('cpu')
    ):
        super().__init__(device=device)
        self._channels = channels

        self.conv1 = nn.Conv2d(self.channels * 3, 128, 1)
        self.conv2 = nn.Conv2d(128, self.channels, 1)

        self.sobol_x_kernel = sobol_x.to(self.device).view(1, 1, 3, 3).repeat(self.channels, 1, 1, 1)
        self.sobol_y_kernel = sobol_y.to(self.device).view(1, 1, 3, 3).repeat(self.channels, 1, 1, 1)

        self.to(self.device)

    def perception(self, state):
        sobol_x = F.conv2d(state, self.sobol_x_kernel, groups=self.channels, padding=1)
        sobol_y = F.conv2d(state, self.sobol_y_kernel, groups=self.channels, padding=1)
        return torch.cat((state, sobol_x, sobol_y), 1)

    def forward(self, state):
        perception = self.perception(state)

        ds = F.relu(self.conv1(perception))
        ds = self.conv2(ds)

        # simulate async update
        random_mask = (torch.rand((state.shape[0], 1, state.shape[2], state.shape[3]), device=self.device) < .5)
        ds = ds * random_mask

        # alive masking
        alive = F.max_pool2d(state[:, 3, :, :], 3, stride=1, padding=1) > .1
        alive = torch.unsqueeze(alive, 1)
        next_state = state + ds
        next_state = next_state * alive

        self._last_delta_state = ds

        return next_state

    @property
    def channels(self) -> int:
        return self._channels

    def reset(self):
        def reset_conv2d(layer):
            n = layer.in_channels
            for k in layer.kernel_size:
                n *= k
            stdv = 1. / torch.sqrt(n)
            layer.weight.data.uniform_(-stdv, stdv)
            if layer.bias is not None:
                layer.bias.data.uniform_(-stdv, stdv)

        reset_conv2d(self.conv1)
        reset_conv2d(self.conv2)
        self.conv2.weight.data.fill_(0.0)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        # cfg["channels"] = self.channels  # Already set in super()
        cfg["model"] = {
            "channels": 128
        }

        return cfg



import datetime
import json
import logging
import re
import tempfile
import warnings
from pathlib import Path
from typing import Optional, Any

import torch
import wandb
from omegaconf import DictConfig
from wandb.apis.public import Files, File, Run
from wandb.util import download_file_from_url

from ncalib.utils.defaults import DEFAULT_DEVICE


def get_latest_model_from_wandb(files: Files) -> Optional[File]:
    logger = logging.getLogger("utils:.wandb:get_latest_model_from_wandb")
    model_files = []
    for file in files:
        assert isinstance(file, File)
        if not file.name.startswith("models"):
            continue

        model_files.append(file)

    logger.debug(f"Found {len(model_files)} models out of {len(files)} files")
    if len(model_files) == 0:
        return None

    return sorted(model_files, key=lambda f: f.name)[-1]


def get_wandb_model(run: Run, *, model_epoch=Optional[int]) -> Optional[dict]:
    pattern = re.compile("model-(\\d+)\\.(state_dict|nca)")

    matched_model_file = None
    current_highest_model = 0

    for file in run.files():
        assert isinstance(file, File)

        result = pattern.findall(file.name)
        if len(result) == 0:
            continue
        epoch_str, file_extension = result[0]
        epoch = int(epoch_str)
        if epoch == model_epoch:
            matched_model_file = file
            break
        elif model_epoch is None and epoch > current_highest_model:
            current_highest_model = epoch
            matched_model_file = file

    if matched_model_file is None:
        warnings.warn(f"No model found (model_epoch={model_epoch})!")
        return None

    tmp_path = Path(tempfile.gettempdir()) / f"{tempfile.gettempprefix()}.{file_extension}"
    download_file_from_url(str(tmp_path), matched_model_file.url, api_key=wandb.Api().api_key)

    full_config = torch.load(tmp_path)
    tmp_path.unlink()  # Delete temp file
    return full_config["state_dict"]


def get_target_file(files: Files) -> Optional[File]:
    logger = logging.getLogger("utils:.wandb:get_target_file")
    for file in files:
        assert isinstance(file, File)
        if file.name.startswith("media/images/target"):
            return file
    logger.error("Could not find a target!")
    return None



def _get_run(run: Optional[Run]) -> Run:
    if run is None:
        if wandb.run is None:
            raise ValueError("No run provided and no run initialized")
        run = wandb.run
    return run


def load_model_from_wandb_run(run: Run = None, *, device=DEFAULT_DEVICE) -> "NCA":
    from hydra.utils import instantiate
    # Take currently logged in run, if run is None
    logger = logging.getLogger("utils:.wandb:load_model_from_wandb_run")
    run = _get_run(run)

    logger.info(f"Loading NCA for run {run}")
    # Build NCA
    wandb_config = load_wandb_hydra_config(run)

    nca = instantiate(wandb_config.algorithm.nca, device=device)
    logger.debug(f"NCA: {nca}")

    # Load model
    files = run.files()
    latest_model = get_latest_model_from_wandb(files)
    if latest_model is None:
        logger.warning(f"No model found for Run {run.id}!")
    else:
        model_path = f"tmp_model_{datetime.datetime.now():%Y%m%d_%H%M%S_%f}.state_dict"
        logger.debug(f"Download {latest_model.id} to {model_path} (Size: {latest_model.size})")
        download_file_from_url(model_path, latest_model.url, api_key=wandb.Api().api_key)

        logger.debug(f"Download complete. Load model.")
        nca.load_state_dict(torch.load(model_path))
        Path(model_path).unlink()  # Delete temp file
    return nca


def get_wandb_values(cfg: Any) -> Any:
    """
    Wandb Config structure:
        key: {"value": value, "desc": ...}

    :return dict with key: value
    """
    if not isinstance(cfg, dict):
        return cfg

    new_cfg = {}
    for key in cfg:
        new_cfg[key] = cfg[key]["value"]

    return new_cfg


def load_wandb_hydra_config(run: Run, prefix="hydra_") -> DictConfig:
    """
    1. Remove prefix from keys
    2. Set value of cfg to "value". Wandb-Configs has structure key: {"value": value, "...": ...}
    """
    cfg = json.loads(run.json_config)

    new_cfg = {}
    for key, value in get_wandb_values(cfg).items():
        if key.startswith(prefix):
            new_cfg[key.replace(prefix, "", 1)] = value

    return DictConfig(new_cfg)


def load_wandb_config(run: Run) -> dict[str, Any]:
    return get_wandb_values(json.loads(run.json_config))



import warnings
from typing import Any
from ncalib.utils.torch_helper import weight_init

import torch
import torch.nn.functional as F
from torch import nn

from ncalib.nca.models import NCA_Model



class BasicNCAModel(NCA_Model):
    def __init__(self, input_size: int, output_size: int, hidden_channels: int = 128):
        super().__init__(input_size, output_size)
        warnings.warn(f"`{self.__class__.__name__}` is a legacy model that is kept only for backwards compatibility reasons. Please use `nca.models.linear.LinearNCAModel` instead!")


        self.conv1 = nn.Conv2d(input_size, hidden_channels, 1)
        self.conv2 = nn.Conv2d(hidden_channels, output_size, 1)
        self.hidden_channels = hidden_channels

    def forward(self, perception: list[torch.Tensor], **kwargs) -> torch.Tensor:
        total_perception = torch.cat(perception, 1)
        delta_state = F.relu(self.conv1(total_perception))
        delta_state = self.conv2(delta_state)

        return delta_state

    def reset(self):
        self.apply(weight_init)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["hidden_channels"] = self.hidden_channels
        return cfg



import abc
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Optional

import cv2
import numpy as np
import torch

from ncalib import NCA_Base
from ncalib.visualization import np_make_grid
from ncalib.visualization.gui.brushes import GuiBrush
from ncalib.visualization.state_visualizer import StateVisualizer


@dataclass
class MouseState:
    _original_x: float = 0
    _original_y: float = 0
    x: float = 0
    y: float = 0
    batch: int = 0
    left_down: bool = False
    right_down: bool = False


class GUIWindow(NCA_Base, abc.ABC):
    def __init__(self):
        super().__init__()
        self.cv2_windows = {}

    def __add__(self, other):
        # Prepare each state updater so that each is an iterable of simple state updaters
        if isinstance(self, MultipleWindows):
            windows1 = self.windows
        else:
            windows1 = [self]

        if isinstance(other, MultipleWindows):
            windows2 = other.windows
        else:
            windows2 = [other]

        # Combine State Updaters
        return MultipleWindows(*windows1, *windows2)

    @abc.abstractmethod
    def update_windows(self):
        pass

    @abc.abstractmethod
    def update_state(self, state: torch.Tensor, step: int):
        pass

    def init_state(self, state: torch.Tensor):
        self.update_state(state, step=0)

    def handle_key(self, key: int):
        pass

    def modify_state(self, state: torch.Tensor) -> torch.Tensor:
        return state


class VisualizingWindow(GUIWindow, abc.ABC):
    def __init__(self, visualizer: StateVisualizer, *, brush: Optional[GuiBrush] = None):
        super().__init__()
        self.visualizer = visualizer
        self.brush = brush
        self.step = 0

        # name -> np.ndarray[B x 3 x W x H]
        self.last_visualization: dict[str, np.ndarray] = {}
        self.mouse_state: dict[str, MouseState] = {}

    def update_state(self, state: torch.Tensor, step: int):
        self.step = step
        for name, images in self.visualizer(state).items():
            B, W, H, C = images.shape
            self.last_visualization[name] = images
            if C == 3 or C == 4:  # SWAP RGB to BGR
                self.last_visualization[name][:, :, :, [0, 2]] = images[:, :, :, [2, 0]]

    def _initialize_window(self, window_name: str):
        try:
            # Destroy old window if it exists
            cv2.destroyWindow(window_name)
        except cv2.error:
            pass

        self.mouse_state[window_name] = MouseState()
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)
        cv2.setMouseCallback(window_name, partial(self.mouse_callback, window_name=window_name))

    def mouse_callback(self, event, x, y, flags, param, *, window_name):
        self.mouse_state[window_name]._original_x = x
        self.mouse_state[window_name]._original_y = y
        image_x, image_y, batch = self.parse_mouse_position(x, y, window_name)
        self.mouse_state[window_name].x = image_x
        self.mouse_state[window_name].y = image_y
        self.mouse_state[window_name].batch = batch
        if event == cv2.EVENT_LBUTTONDOWN:
            self.mouse_state[window_name].left_down = True
        elif event == cv2.EVENT_LBUTTONUP:
            self.mouse_state[window_name].left_down = False
        # elif event == cv2.EVENT_RBUTTONDown:
        #     self.mouse[batch].right_down = True
        # elif event == cv2.EVENT_RBUTTONUP:
        #     self.mouse[batch].right_down = False

    def modify_state(self, state: torch.Tensor) -> torch.Tensor:
        if self.brush is None:
            return state

        for mouse in self.mouse_state.values():
            state = self.brush.modify_state(state, mouse=mouse)
        return state

    def parse_mouse_position(
            self,
            x: float,
            y: float,
            window_name: str
    ) -> tuple[Optional[float], Optional[float], Optional[int]]:
        """
        :return: image_x, image_y, batch
        """
        return None, None, None


class MultipleWindows(GUIWindow):
    def __init__(self, *windows: GUIWindow):
        super().__init__()
        self.windows = tuple(windows)

    def update_windows(self):
        for window in self.windows:
            window.update_windows()

    def init_state(self, state: torch.Tensor):
        for window in self.windows:
            window.init_state(state)

    def handle_key(self, key: int):
        for window in self.windows:
            window.handle_key(key)

    def update_state(self, state: torch.Tensor, step: int):
        for window in self.windows:
            window.update_state(state, step=step)

    def modify_state(self, state: torch.Tensor) -> torch.Tensor:
        for window in self.windows:
            state = window.modify_state(state)
        return state


class SliderBatchWindow(VisualizingWindow):
    def __init__(self, visualizer: StateVisualizer, *, brush: Optional[GuiBrush] = None):
        super().__init__(visualizer, brush=brush)
        self.batch_selected = 0

    def init_state(self, state: torch.Tensor):
        super().init_state(state)
        for window_name, images in self.last_visualization.items():
            batch_size = len(images)
            self._initialize_window(window_name)
            if batch_size > 1:
                cv2.createTrackbar("Batch", window_name, 0, batch_size - 1, self._batch_slider_changed)

    def _batch_slider_changed(self, value):
        self.batch_selected = value - 1

    def update_windows(self):
        for name, images in self.last_visualization.items():
            cv2.imshow(name, images[self.batch_selected])

    def parse_mouse_position(
            self,
            x: float,
            y: float,
            window_name: str
    ) -> tuple[Optional[float], Optional[float], Optional[int]]:
        """
        :return: image_x, image_y, batch
        """
        return x, y, self.batch_selected


class MergeBatchWindow(VisualizingWindow):
    def __init__(
            self,
            visualizer: StateVisualizer,
            *,
            brush: Optional[GuiBrush] = None,
            padding: int = 2,
            batch_width: Optional[int] = None,
            save_image_folder: Optional[Path | str] = None
    ):
        super().__init__(visualizer, brush=brush)
        self.nrow = batch_width
        self.batch_selected = 0
        self.padding = padding
        self.save_image_folder = Path(save_image_folder) if save_image_folder is not None else None

    def init_state(self, state: torch.Tensor):
        super().init_state(state)
        for window_name, images in self.last_visualization.items():
            batch_size = len(images)
            self._initialize_window(window_name)
            if self.nrow is None:
                self.nrow = int(batch_size ** 0.5 + 0.49)
        self.update_windows()

    def update_windows(self):
        for name, images in self.last_visualization.items():
            merged_image = np_make_grid(
                images,
                nrow=self.nrow,
                pad_value=255,
                padding=self.padding
            )
            cv2.imshow(name, merged_image)

            if self.save_image_folder is not None:
                self.save_image_folder.mkdir(exist_ok=True, parents=True)
                cv2.imwrite(str(self.save_image_folder / f"{name}_{self.step}.png"), merged_image)

    def parse_mouse_position(
            self,
            x: float,
            y: float,
            window_name: str
    ) -> tuple[Optional[float], Optional[float], Optional[int]]:
        """
        :return: image_x, image_y, batch
        """
        B, W, H, C = self.last_visualization[window_name].shape
        xmaps = min(self.nrow, B)
        height, width = int(H + self.padding), int(W + self.padding)
        batch = max(min(int((y // height) * xmaps + x // width), B - 1), 0)
        return x % width, y % height, batch


class SeparateWindows(VisualizingWindow):
    def __init__(self, visualizer: StateVisualizer, *, brush: Optional[GuiBrush] = None):
        super().__init__(visualizer, brush=brush)
        self.batch_selected = 0

    def init_state(self, state: torch.Tensor):
        super().init_state(state)
        for single_window_name, images in self.last_visualization.items():
            batch_size = len(images)
            if batch_size == 1:
                window_names = [single_window_name]
            else:
                window_names = [self.window_name(single_window_name, i) for i in range(batch_size)]
            for name in window_names:
                self._initialize_window(name)

    def update_windows(self):
        for name, images in self.last_visualization:
            for i, image in enumerate(images):
                cv2.imshow(self.window_name(name, i), image)

    @staticmethod
    def window_name(name: str, i: int) -> str:
        return f"{name}-{i}"

    def parse_mouse_position(
            self,
            x: float,
            y: float,
            window_name: str
    ) -> tuple[Optional[float], Optional[float], Optional[int]]:
        """
        :return: image_x, image_y, batch
        """
        batch = int(window_name.split("-")[1])
        return x, y, batch



from typing import Optional

import numpy as np
import torch
import torchvision.utils
from jaxtyping import Integer, Float
import matplotlib.pyplot as plt

from ncalib.visualization.state_visualizer import BatchedStateVisualizer


class RGBVisualizer(BatchedStateVisualizer):
    def __init__(
            self,
            *,
            red_channel: int = 0,
            green_channel: int = 1,
            blue_channel: int = 2,
            alpha_channel: Optional[int] = None,
            batch_slice=slice(None, None, None),
            name: str = "RGB"
    ):
        super().__init__(name=name, batch_slice=batch_slice)
        self.red_channel = red_channel
        self.green_channel = green_channel
        self.blue_channel = blue_channel
        self.alpha_channel = alpha_channel

    def _batch_to_visualization(self, batch: torch.Tensor) -> dict[str, torch.Tensor]:
        """
        :param batch: Tensor[B x C x W x H]
        :return: dict[name -> Tensor[B x 3/4 x W x H] with float ranges 0..1 in RGB, RGBA
        """
        channels = [self.red_channel, self.green_channel, self.blue_channel]
        if self.alpha_channel is not None:
            channels += [self.alpha_channel]
        rgb_channels = batch[:, channels]
        return {
            self.name: rgb_channels
        }


class SeparateChannelVisualizer(BatchedStateVisualizer):
    def __init__(
            self,
            channels: slice = slice(None, None, None),
            *,
            cmap: Optional[str] = 'viridis',
            vmin: Optional[float] = None,
            vmax: Optional[float] = None,
            name: str = "Channels",
            batch_slice=slice(None, None, None),
    ):
        super().__init__(name=name, batch_slice=batch_slice)

        self.channels = channels
        self.cmap = cmap
        self.vmin = vmin
        self.vmax = vmax

    def _batch_to_visualization(
            self,
            batch: Float[torch.Tensor, "B C W H"]
    ) -> dict[str, Integer[np.ndarray, "B 3 W' H'"] | Float[torch.Tensor, "B 1 W' H'"]]:
        """
        :param batch: Tensor[B x C x W x H]
        :return: dict[name -> Tensor[B x 1 x W' x H'] with float ranges 0..1 in Gray
        """
        batched_results = []
        channels = batch.shape[1]
        n_rows = int(channels ** 0.5)
        for image in batch[:, self.channels]:
            # image: Tensor[C x W x H]
            image_unsqueezed = image.unsqueeze(1)
            # image_unsqueezed: Tensor [C x 1 x W x H]
            grid_image = torchvision.utils.make_grid(image_unsqueezed, nrow=n_rows)[:1]  # Helper creates RGB image
            # grid_image: Tensor[1 x W' x H']
            batched_results.append(grid_image)
        return {self.name: self._apply_cmap(torch.stack(batched_results, dim=0))}

    def _apply_cmap(self, batched_results: Float[torch.Tensor, "B 1 W H"]) -> Integer[np.ndarray, "B 3 W H"] | Float[
        torch.Tensor, "B 1 W H"]:
        if self.cmap is None:
            return batched_results

        # Normalize the tensor values
        vmin = batched_results.min() if self.vmin is None else self.vmin
        vmax = batched_results.max() if self.vmax is None else self.vmax
        batched_results_normalized = (batched_results - vmin) / (vmax - vmin)

        # Get the colormap
        cmap = plt.get_cmap(self.cmap)

        # Apply the colormap
        tensor_colored = []
        for t in batched_results_normalized:
            # Convert to numpy and apply colormap
            t_np = t.squeeze().detach().cpu().numpy()
            t_colored_np = cmap(t_np)[:, :, :3]  # Get RGB channels

            tensor_colored.append(t_colored_np)

        tensor_colored = np.stack(tensor_colored)
        return (tensor_colored * 255).astype(np.uint8)


class DeltaMagnitudeVisualizer(BatchedStateVisualizer):
    def __init__(
            self,
            *,
            channels: slice = slice(None, None, None),
            batch_slice=slice(None, None, None),
            normalize: bool = True,
            name: str = "Delta Magnitude"
    ):
        super().__init__(name=name, batch_slice=batch_slice)
        self.channels = channels

        self.normalize = normalize
        self._old_batch = None

    def _batch_to_visualization(self, batch: torch.Tensor) -> dict[str, torch.Tensor]:
        """
        :param batch: Tensor[B x C x W x H]
        :return: dict[name -> Tensor[B x 1 x W x H] with float ranges 0..1 in Gray

        """
        if self._old_batch is None:
            self._old_batch = torch.zeros_like(batch)

        magnitudes = torch.sqrt(torch.sum((batch[:, self.channels] - self._old_batch) ** 2, dim=1, keepdim=True))
        if self.normalize:
            magnitudes = (magnitudes - magnitudes.min()) / (magnitudes.max() - magnitudes.min())
        self._old_batch = batch
        return {self.name: magnitudes}



from typing import Optional

import torch

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.utils.defaults import DEFAULT_DEVICE


class ConstantFillSeedFactory(NCA_SeedFactory):
    def __init__(
            self,
            width: int = 64,
            channels: int = 16,
            height: Optional[int] = None,
            fill_value: float = 0
    ):
        super().__init__(width=width, height=height, channels=channels)
        self.fill_value = fill_value

    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE) -> torch.Tensor:
        seed = torch.ones([n, self.channels, self.width, self.height], dtype=torch.float32, device=device)
        return seed * self.fill_value

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["fill_value"] = self.fill_value
        return cfg


class NormalSeedFactory(NCA_SeedFactory):
    def __init__(
            self,
            width: int = 64,
            channels: int = 16,
            height: Optional[int] = None,
            visible_channel: int = 3
    ):
        super().__init__(width=width, height=height, channels=channels)
        self.visible_channel = visible_channel

    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE) -> torch.Tensor:
        seed = torch.zeros([n, self.channels, self.width, self.height], dtype=torch.float32, device=device)
        seed[:, self.visible_channel:, self.width // 2, self.height // 2] = 1.0
        return seed

    @classmethod
    def for_target(cls, target: "NCA_Target", nca: "NCA") -> "NormalSeedFactory":
        return NormalSeedFactory(
            width=target.width,
            height=target.height,
            channels=nca.channels,
        )

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["visible_channel"] = self.visible_channel
        return cfg


class SpreadSeedFactory(NCA_SeedFactory):
    def __init__(
            self,
            width: int = 64,
            channels: int = 16,
            height: Optional[int] = None,
            visible_channel: int = 3,
            *,
            seed=0
    ):
        super().__init__(width=width, height=height, channels=channels)
        self.visible_channel = visible_channel
        with torch.random.fork_rng():
            torch.random.manual_seed(seed)
            self._seed = torch.rand([self.channels, self.width, self.height], dtype=torch.float32)
            self._seed[:visible_channel] = 0

    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE) -> torch.Tensor:
        seed = self._seed.clone()
        return seed.to(device).repeat([n, 1, 1, 1])

    @classmethod
    def for_target(cls, target: "NCA_Target", nca: "NCA") -> "NormalSeedFactory":
        return NormalSeedFactory(
            width=target.width,
            height=target.height,
            channels=nca.channels,
        )

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["visible_channel"] = self.visible_channel
        return cfg





import dataclasses
import math
from dataclasses import dataclass

import torch
from PIL import Image

from ncalib.utils.image import RGBAImage
from ncalib.utils.utils import full_class_name


def rgba_loader(path):
    with open(path, "rb") as f:
        img = Image.open(f)
        return img.convert("RGBA")

def rgba_loader_with_aux(path):
    image = rgba_loader(path)
    return image


class NoColorChannelSelectedException(Exception):
    pass


@dataclass(frozen=True)
class ColorClass:
    def __call__(self, rgba_image: RGBAImage) -> torch.Tensor:
        result_list = []
        for channel_name in self.__class__.__dataclass_fields__.keys():
            # channel_value can be either ColorClass or bool
            channel_value = getattr(self, channel_name)

            if isinstance(channel_value, ColorClass):  # Case recursive ColorClass
                try:
                    channels = channel_value(rgba_image)
                except NoColorChannelSelectedException:
                    continue

                for channel in channels:
                    result_list.append(channel)
            elif channel_value is True:  # Case True
                channel_conversion = getattr(self, f"_{channel_name}", None)
                if channel_conversion is None:
                    raise NotImplementedError(f"{self.__class__} has no converter for channel {channel_name}")
                channel = channel_conversion(rgba_image)
                result_list.append(channel)
            elif channel_value is not False:  # Unknown type
                raise TypeError(f"Cannot identify value {channel_value} for {channel_name}")

        if len(result_list) == 0:
            raise NoColorChannelSelectedException()
        return torch.stack(result_list)

    def __len__(self) -> int:
        return len(self.names)  # Shortcut... Might be more efficient if done properly

    @property
    def names(self) -> list[str]:
        names = []
        for channel_name in self.__class__.__dataclass_fields__.keys():
            # channel_value can be either ColorClass or bool
            channel_value = getattr(self, channel_name)

            if isinstance(channel_value, ColorClass):  # Case recursive ColorClass
                names += channel_value.names
            elif channel_value is True:  # Case True
                names.append(f"{self.__class__.__name__}.{channel_name}")
            elif channel_value is not False:  # Unknown type
                raise TypeError(f"Cannot identify value {channel_value} for {channel_name}")

        return names

    def to_config_dict(self):
        cfg = {
            "_target_": full_class_name(self)
        }
        cfg.update(dataclasses.asdict(self))
        return cfg


@dataclass(frozen=True)
class RGBA(ColorClass):
    red: bool = False
    green: bool = False
    blue: bool = False
    alpha: bool = False

    @classmethod
    def _red(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return rgba_image.rgba_red

    @classmethod
    def _green(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return rgba_image.rgba_green

    @classmethod
    def _blue(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return rgba_image.rgba_blue

    @classmethod
    def _alpha(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return rgba_image.rgba_alpha


@dataclass(frozen=True)
class HSV(ColorClass):
    """ https://en.wikipedia.org/wiki/HSL_and_HSV """
    hue: bool = False
    saturation: bool = False
    value: bool = False

    @classmethod
    def _hue(cls, rgba_image: RGBAImage) -> torch.Tensor:
        hue = torch.zeros((rgba_image.width, rgba_image.height))
        case_r = rgba_image.rgb_max == rgba_image.rgba_red
        case_g = rgba_image.rgb_max == rgba_image.rgba_green
        case_b = rgba_image.rgb_max == rgba_image.rgba_blue
        hue[case_r] = torch.remainder((rgba_image.rgba_green - rgba_image.rgba_blue) / rgba_image.rgb_delta, 6)[case_r]
        hue[case_g] = ((rgba_image.rgba_blue - rgba_image.rgba_red) / rgba_image.rgb_delta + 2)[case_g]
        hue[case_b] = ((rgba_image.rgba_red - rgba_image.rgba_green) / rgba_image.rgb_delta + 4)[case_b]
        return torch.nan_to_num(hue, 0) / 6

    @classmethod
    def _saturation(cls, rgba_image: RGBAImage) -> torch.Tensor:
        values = rgba_image.rgb_delta / HSV._value(rgba_image)
        return torch.clip(torch.nan_to_num(values, 0), min=0, max=1)

    @classmethod
    def _value(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return rgba_image.rgb_max


@dataclass(frozen=True)
class HSL(ColorClass):
    """ https://en.wikipedia.org/wiki/HSL_and_HSV """
    hue: bool = False
    saturation: bool = False
    luminance: bool = False

    @classmethod
    def _hue(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return HSV._hue(rgba_image)

    @classmethod
    def _saturation(cls, rgba_image: RGBAImage) -> torch.Tensor:
        values = rgba_image.rgb_delta / (1 - torch.abs(2 * HSL._luminance(rgba_image) - 1))
        return torch.clip(torch.nan_to_num(values, 0), min=0, max=1)

    @classmethod
    def _luminance(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return (rgba_image.rgb_max + rgba_image.rgb_min) / 2


@dataclass(frozen=True)
class HSI(ColorClass):
    hue: bool = False
    saturation: bool = False
    intensity: bool = False

    @classmethod
    def _hue(cls, rgba_image: RGBAImage) -> torch.Tensor:
        """ https://en.wikipedia.org/wiki/Hue#Defining_hue_in_terms_of_RGB """
        alpha = 2 * rgba_image.rgba_red - rgba_image.rgba_green - rgba_image.rgba_blue
        beta = math.sqrt(3) * (rgba_image.rgba_green - rgba_image.rgba_blue)
        values = torch.atan2(beta, alpha)

        # Normalize between 0..1
        return (values + math.pi) / (2 * math.pi)

    @classmethod
    def _saturation(cls, rgba_image: RGBAImage) -> torch.Tensor:
        res = 1 - rgba_image.rgb_min / cls._intensity(rgba_image)
        return torch.clip(torch.nan_to_num(res, 0), min=0)

    @classmethod
    def _intensity(cls, rgba_image: RGBAImage) -> torch.Tensor:
        """ Unweighted Gray """
        return torch.mean(rgba_image.img[:3], dim=0)  # "Gray"


@dataclass(frozen=True)
class CMYK(ColorClass):
    cyan: bool = False
    magenta: bool = False
    yellow: bool = False
    key: bool = False

    @classmethod
    def _key(cls, rgba_image: RGBAImage):
        return 1 - rgba_image.rgb_max

    @classmethod
    def _cyan(cls, rgba_image: RGBAImage) -> torch.Tensor:
        res = (rgba_image.rgb_max - rgba_image.rgba_red) / rgba_image.rgb_max
        return torch.nan_to_num(res, 0)

    @classmethod
    def _magenta(cls, rgba_image: RGBAImage) -> torch.Tensor:
        res = (rgba_image.rgb_max - rgba_image.rgba_green) / rgba_image.rgb_max
        return torch.nan_to_num(res, 0)

    @classmethod
    def _yellow(cls, rgba_image: RGBAImage) -> torch.Tensor:
        res = (rgba_image.rgb_max - rgba_image.rgba_blue) / rgba_image.rgb_max
        return torch.nan_to_num(res, 0)


@dataclass(frozen=True)
class YUV(ColorClass):
    """ https://en.wikipedia.org/wiki/YUV """
    _V_max = 0.615
    _U_max = 0.436
    y: bool = False
    u: bool = False
    v: bool = False

    @classmethod
    def _y(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return 0.299 * rgba_image.rgba_red + 0.587 * rgba_image.rgba_green + 0.114 * rgba_image.rgba_blue

    @classmethod
    def _u(cls, rgba_image: RGBAImage) -> torch.Tensor:
        values = -0.14713 * rgba_image.rgba_red + -0.28886 * rgba_image.rgba_green + 0.436 * rgba_image.rgba_blue
        return (values + cls._U_max) / (2 * cls._U_max)

    @classmethod
    def _v(cls, rgba_image: RGBAImage) -> torch.Tensor:
        values = 0.615 * rgba_image.rgba_red + -0.51499 * rgba_image.rgba_green + -0.10001 * rgba_image.rgba_blue
        return (values + cls._V_max) / (2 * cls._V_max)


@dataclass(frozen=True)
class RG_Chromaticity(ColorClass):
    """ https://en.wikipedia.org/wiki/Rg_chromaticity """
    red: bool = False
    green: bool = False
    blue: bool = False

    @classmethod
    def _red(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return torch.nan_to_num(rgba_image.rgba_red / torch.sum(rgba_image.img[:3], dim=0), 0)

    @classmethod
    def _green(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return torch.nan_to_num(rgba_image.rgba_green / torch.sum(rgba_image.img[:3], dim=0), 0)

    @classmethod
    def _blue(cls, rgba_image: RGBAImage) -> torch.Tensor:
        # Edge case: B' = 1 - R' - G' -> if R + G + B == 0 -> B' = 1
        return torch.nan_to_num(rgba_image.rgba_blue / torch.sum(rgba_image.img[:3], dim=0), 1)


@dataclass(frozen=True)
class TSL(ColorClass):
    """ https://en.wikipedia.org/wiki/TSL_color_space """
    tint: bool = False
    saturation: bool = False
    lightness: bool = False

    @classmethod
    def _tint(cls, rgba_image: RGBAImage) -> torch.Tensor:
        r_ = RG_Chromaticity._red(rgba_image) - 1 / 3
        g_ = RG_Chromaticity._green(rgba_image) - 1 / 3
        tint = torch.zeros((rgba_image.width, rgba_image.height))

        case_1 = g_ > 0
        case_2 = g_ < 0
        tint[case_1] = (torch.arctan(r_ / g_ + 1 / 4) / (2 * torch.pi))[case_1]
        tint[case_2] = (torch.arctan(r_ / g_ + 3 / 4) / (2 * torch.pi))[case_2]
        return (torch.nan_to_num(tint, 0) + 0.25) * 2  # / (2 * math.pi)

    @classmethod
    def _saturation(cls, rgba_image: RGBAImage) -> torch.Tensor:
        r_ = rgba_image.rgba_red - 1 / 3
        g_ = rgba_image.rgba_red - 1 / 3
        return torch.sqrt(9 / 5 * (r_ ** 2 + g_ ** 2)) / math.sqrt(1.6)

    @classmethod
    def _lightness(cls, rgba_image: RGBAImage) -> torch.Tensor:
        return 0.299 * rgba_image.rgba_red + 0.587 * rgba_image.rgba_green + 0.114 * rgba_image.rgba_blue


@dataclass(frozen=True)
class ColorSpace(ColorClass):
    rgba: RGBA = RGBA()
    hsv: HSV = HSV()
    hsl: HSL = HSL()
    hsi: HSI = HSI()
    cmyk: CMYK = CMYK()
    yuv: YUV = YUV()
    rg_chromaticity: RG_Chromaticity = RG_Chromaticity()
    tsl: TSL = TSL()



from typing import Any

import torch

from ncalib.trainer.batcher.pool import PoolEvictionPolicy
from ncalib.trainer.dataset import DataSample


class LossSortEvictionPolicy(PoolEvictionPolicy):
    def __init__(self, *, logarithmic=True):
        super().__init__()
        self.logarithmic = logarithmic

    def score_batch(self, data_sample: DataSample, losses: torch.Tensor) -> torch.Tensor:
        scores = losses
        if self.logarithmic:
            scores = torch.log(scores)

        scores_min = scores.min()
        scores_max = scores.max()
        if scores_max - scores_min != 0:
            scores = (scores - scores_min) / (scores_max - scores_min)
            return scores

        return torch.zeros_like(scores)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["logarithmic"] = self.logarithmic
        return cfg


class MeanLossEvictionPolicy(PoolEvictionPolicy):
    def __init__(self, *, logarithmic=True):
        super().__init__()
        self.logarithmic = logarithmic

    def score_batch(self, data_sample: DataSample, losses: torch.Tensor) -> torch.Tensor:
        scores = losses
        if self.logarithmic:
            scores = torch.log(scores)

        scores = torch.abs(scores - scores.mean())
        max_score = torch.max(scores)
        if max_score != 0:
            scores = scores / torch.max(scores)
        return scores

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["logarithmic"] = self.logarithmic
        return cfg



from typing import Optional, Any

import torch

from ncalib.nca.state_updater import NCA_StateUpdater


class AddNormalNoiseStateUpdater(NCA_StateUpdater):
    def __init__(self, mu: float = 0, std: float = 0.02, *, seed: Optional[int] = None):
        super().__init__()
        self.mu = mu
        self.std = std
        self.rng = None
        self.seed = seed

    def forward(
            self,
            state: torch.Tensor,
            delta_state: torch.Tensor,
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        if self.rng is None:
            self.rng = torch.Generator(device=state.device)
            if self.seed is not None:
                self.rng.manual_seed(self.seed)
            else:
                self.rng.seed()
        noise = torch.normal(self.mu, self.std, delta_state.shape, generator=self.rng, device=state.device)
        return state, delta_state + noise

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["mu"] = self.mu
        cfg["std"] = self.std
        return cfg



import abc
from typing import Callable, Any

import torch

from ncalib.trainer.dataset import NCA_Dataset, DataSample
from ncalib.trainer.loss_function import LossResult
from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.utils import NCA_Base


class NCA_Batcher(NCA_Base, abc.ABC):
    def __init__(
            self,
            dataset: NCA_Dataset,
            batch_size: int,
            num_workers: int = 0,
            collate_fn: Callable = DataSample.merge_batch,
            device: torch.device | str = DEFAULT_DEVICE
    ):
        super().__init__()
        self.collate_fn = collate_fn
        self.batch_size = batch_size
        self.dataset = dataset
        self.num_workers = num_workers
        self.device = device

    @abc.abstractmethod
    def next_batch(self) -> DataSample:
        pass

    def init(self):
        pass

    def feedback(self, data_sample: DataSample, new_state: torch.Tensor, loss: LossResult):
        pass

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["dataset"] = self.dataset.to_config_dict()
        cfg["batch_size"] = self.batch_size
        cfg["num_workers"] = self.num_workers
        return cfg



from typing import Optional, Any

from torch.utils.data import Dataset, DataLoader

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.trainer.dataset import DataSample, NCA_SteppedDataset
from ncalib.utils.torch_helper import EndlessIterableDataset
from ncalib.utils.utils import dataset_to_config


class ImageTrainingDataset(NCA_SteppedDataset):
    def __init__(
            self,
            seed_factory: NCA_SeedFactory,
            target: Dataset,
            *,
            step_range: tuple[int, int],
            training_regenerate=False,
            seed: Optional[int] = None,
    ):
        super().__init__(step_range, seed=seed)
        self.seed_factory = seed_factory
        self.target = target
        self.target_dataloader = iter(
            DataLoader(
                EndlessIterableDataset(target, generator=self.rng, shuffle=True),
            )
        )
        self.training_regenerate = training_regenerate

    def __call__(self) -> DataSample:
        num_step = self.generate_n_steps()

        state = self.seed_factory(1, device="cpu")
        target = next(self.target_dataloader)

        return DataSample(state, target, num_step, training_regenerate=[self.training_regenerate], batch_size=[1])

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["seed_factory"] = self.seed_factory.to_config_dict()
        cfg["target"] = dataset_to_config(self.target)
        cfg["training_regenerate"] = self.training_regenerate
        return cfg





from typing import Optional, Protocol, Any

import torch
import torch.nn.functional as F
from jaxtyping import Float

from ncalib import full_class_name
from ncalib.trainer.loss_function import NamedLossFunction
from ncalib.utils.utils import slice_to_config


class _LossFnProtocol(Protocol):
    def __call__(self, state: torch.Tensor, target: torch.Tensor, reduction: Optional[str], **kwargs) -> torch.Tensor:
        pass


class _ReductionFnProtocol(Protocol):
    def __call__(self, values: torch.Tensor, *, dim: tuple[int, int, int]) -> torch.Tensor:
        pass


class ImageTargetLoss(NamedLossFunction):
    def __init__(
            self,
            state_channels: slice = slice(0, 3),
            state_weights_channel: Optional[slice | int] = None,
            target_channels: slice = slice(None, None),
            *,
            loss_fn: _LossFnProtocol = F.mse_loss,
            reduction_fn: _ReductionFnProtocol = torch.mean,
            target_position_x: slice = slice(None, None),
            target_position_y: slice = slice(None, None),
            loss_name: str = "rgb",
            state_activation = None
    ):
        super().__init__(loss_name=loss_name)
        self.state_channels = state_channels
        self.target_channels = target_channels
        if isinstance(state_weights_channel, int):
            state_weights_channel = slice(state_weights_channel, state_weights_channel + 1)
        self.state_weights_channel = state_weights_channel
        self.loss_fn = loss_fn
        self.reduction_fn = reduction_fn
        self.target_position_x = target_position_x
        self.target_position_y = target_position_y
        self.state_activation = state_activation

    def _calculate_loss(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B K W H"],
            **kwargs
    ) -> Float[torch.Tensor, "B"]:
        state = state_progression[-1]
        if self.state_weights_channel is None:
            weights = 1
        else:
            weights = torch.mean(torch.abs(state[:, self.state_weights_channel]), dim=1, keepdim=True)
        cropped_state = state[:, self.state_channels, self.target_position_x, self.target_position_y]
        if self.state_activation is not None:
            cropped_state = self.state_activation(cropped_state)

        cellwise_loss = self.loss_fn(cropped_state, target[:, self.target_channels].to(state.device), reduction="none",
                                     **kwargs)
        loss_data = self.reduction_fn(
            cellwise_loss * weights,
            dim=(1, 2, 3)
        )
        return loss_data

    @classmethod
    def create_from_sizes(
            cls,
            seed_size: tuple[int, int],
            target_size: tuple[int, int],
            state_channels: slice = slice(0, 3),
            *,
            loss_fn: _LossFnProtocol = F.mse_loss,
            reduction_fn: _ReductionFnProtocol = torch.mean
    ):
        target_pad_x_start = (seed_size[0] - target_size[0]) // 2
        target_pad_x_end = target_pad_x_start + target_size[0]
        target_pad_y_start = (seed_size[1] - target_size[1]) // 2
        target_pad_y_end = target_pad_y_start + target_size[1]

        return cls(
            state_channels=state_channels,
            loss_fn=loss_fn,
            target_position_x=slice(target_pad_x_start, target_pad_x_end),
            target_position_y=slice(target_pad_y_start, target_pad_y_end),
            reduction_fn=reduction_fn
        )

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["loss_fn"] = full_class_name(self.loss_fn)
        cfg["state_channels"] = slice_to_config(self.state_channels)
        cfg["state_weights_channel"] = slice_to_config(self.state_weights_channel)
        cfg["target_channels"] = slice_to_config(self.target_channels)
        cfg["target_position_x"] = slice_to_config(self.target_position_x)
        cfg["target_position_y"] = slice_to_config(self.target_position_y)
        cfg["state_activation"] = self.state_activation.__name__ if self.state_activation is not None else None
        return cfg



import random
from pathlib import Path
from typing import Optional, Callable, Any

import torch

from ncalib.seed_factory.genome import GenomeSeedFactory
from ncalib.trainer.dataset import DataSample, NCA_SteppedDataset
from ncalib.utils.color_space import rgba_loader
from ncalib.utils.utils import to_binary_list


def load_targets(folder: Path, loader, transform, *, choose_k: Optional[int] = None) -> dict[int, torch.Tensor]:
    res = {}
    for image_path in folder.iterdir():
        if not image_path.is_file():
            continue

        image_id = int(image_path.stem)
        image = loader(image_path)
        image = transform(image)
        res[image_id] = image

    if choose_k is not None:
        keys = list(res.keys())
        random.shuffle(keys)
        for key in keys[choose_k:]:
            res.pop(key)
        assert len(res) == choose_k

    return res


class GenomeDataset(NCA_SteppedDataset):
    def __init__(
            self,
            target_folder: Path | str,
            step_range: tuple[int, int],
            nca_channels: int,
            *,
            seed: Optional[int] = None,
            transform: Optional[Callable] = None,
            loader: Callable[[str], Any] = rgba_loader,
            choose_k: Optional[int] = None,
            deterministic: bool = False,
    ):
        super().__init__(step_range, seed=seed)
        self.target_folder = Path(target_folder)
        self.targets = load_targets(self.target_folder, loader, transform, choose_k=choose_k)
        self.choose_k = choose_k
        self.nca_channels = nca_channels

        # Properties
        self.target_keys = list(self.targets.keys())
        self.n_targets = len(self.targets)
        self.genome_size = len(f"{max(self.target_keys):b}")

        # Seed factory
        self.seed_factory = self.create_seed_factory(deterministic=deterministic)

    def __str__(self):
        return f"{self.__class__.__name__}({self.target_folder.name}, n_targets={self.n_targets}, genome_size={self.genome_size})"

    def create_seed_factory(self, *, deterministic=False) -> GenomeSeedFactory:
        single_target = list(self.targets.values())[0]
        C, W, H = single_target.shape

        return GenomeSeedFactory(
            visible_channels=slice(0, C),
            genome_channels=slice(C, C + self.genome_size),
            channels=self.nca_channels,
            width=W,
            height=H,
            deterministic=deterministic
        )

    def __call__(self) -> DataSample:
        num_step = self.generate_n_steps()

        idx = torch.randint(low=0, high=self.n_targets, size=(1,), generator=self.rng)
        key = self.target_keys[idx]
        target = self.targets[key]
        genome = to_binary_list(key, min_length=self.genome_size)

        state = self.seed_factory(1, genome=genome, device=target.device)

        return DataSample(state, target[None], num_step, training_regenerate=True, batch_size=[1])

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["target_folder"] = str(self.target_folder)
        cfg["choose_k"] = self.choose_k
        return cfg



import warnings
from typing import Optional, Iterable, Literal, Callable

import cv2
import matplotlib.colors as mc
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import torch
from PIL import Image

from ncalib import NCA
from ncalib.utils.utils import length_of_slice
from ncalib.visualization.state_visualizer import StateVisualizer
from ncalib.visualization.state_visualizer.simple import RGBVisualizer

ColorType = str | tuple[float, float, float]


# Helper functions
def get_ax(ax: Optional[plt.Axes]) -> plt.Axes:
    """
    Return input axis. If it is None, return current axis instead (plt.gca())
    """
    if ax is None:
        ax = plt.gca()

    assert isinstance(ax, plt.Axes)
    return ax


def get_color(color: ColorType) -> tuple[float, float, float]:
    """
    Input can be either string (name or hex) or rgb-tuple (scale between 0..1).
    Output is always rgb-tuple scaled between 0..1
    """
    return mc.to_rgb(color)


def clean_state(state: torch.Tensor, *, clamp: bool = False) -> torch.Tensor:
    """
    :param state:  [W x H] or [C x W x H] or [B x C x W x H]
    :param clamp: if True, clamp values between 0..1
    :return: [C x W x H]
    """
    state = state.detach().cpu()
    if clamp:
        state = state.clamp(0, 1)

    if len(state.shape) == 2:  # [W x H]
        return torch.unsqueeze(state, 0)
    elif len(state.shape) == 3:  # [C x W X H]
        return state
    elif len(state.shape) == 4:  # [B x C x W x H]
        return state[0]
    raise ValueError(f"Cannot identify first state from tensor with shape {state.shape}")


def as_quadratic_shape_as_possible(n: int) -> tuple[int, int]:
    w = int(np.ceil(np.sqrt(n)))
    for h in range(1, w + 1):
        if h * w >= n:
            return w, h
    raise ValueError(f"Could not find valid w, h for {n} ({w=})")


# Visualization
def timeline_states_to_rgb_pillow(timeline_states: torch.Tensor, *, channels: slice = slice(0, 3)) -> list[list[Image]]:
    return [
        states_to_rgb_pillow(states, channels=channels)
        for states in timeline_states
    ]


def states_to_rgb_pillow(states: torch.Tensor, *, channels: slice = slice(0, 3)) -> list[Image]:
    """
    :param states: [B x C x W x H]
    :param channels: Used channels for image. Can have length of 1, 3 or 4
    """
    return [
        state_to_rgb_pillow(state, channels=channels)
        for state in states
    ]


def state_to_rgb_pillow(state: torch.Tensor, *, channels: slice = slice(0, 3)) -> Image:
    """
    :param state: [C x W x H] or [1 x C x W x H]
    :param channels: Used channels for image. Can have length of 1, 3 or 4
    """
    state = clean_state(state, clamp=True)[channels]
    C, W, H = state.shape
    processed_state = (state * 255).numpy().astype("uint8").transpose((1, 2, 0))
    if C == 1:
        return Image.fromarray(processed_state[:, :, 0], mode="L")
    elif C == 3:
        mode = "RGB"
    elif C == 4:
        mode = "RGBA"
    else:
        raise ValueError(f"No mode know for {C} channels ({channels=})")

    return Image.fromarray(processed_state, mode=mode)


def states_to_rgb_tensors(states: torch.Tensor, *, channels: slice = slice(0, 3)) -> torch.Tensor:
    """
    :param states: [B x C x W x H]
    :param channels: Used channels for image. Can have length of 1, 3 or 4
    :return [B x W x H x (3, 4)]
    """
    return states[:, channels].detach().clamp(0, 1).permute(0, 2, 3, 1).cpu()


def state_to_bgr_numpy(state: torch.Tensor, *, channels: slice = slice(0, 3)) -> np.ndarray:
    """
    :param state: [C x W x H] or [1 x C x W x H]
    :param channels: Used channels for image. Can have length of 1, 3 or 4
    :return: np.ndarray[W x H x (3 or 4)]
    """
    state = clean_state(state, clamp=True)[channels]
    img = (state * 255).numpy().astype("uint8")  # 0..1 -> 0.255
    img = img[::-1]
    return img.transpose((1, 2, 0))  # [C x W x H] -> [W x H x C]


def channel_to_image_pillow(
        channel: torch.Tensor,
        *,
        colormap: int = cv2.COLORMAP_SUMMER,
        normalization: Literal["arctan", "linear", "clip"] | Callable[[np.ndarray], np.ndarray] = "clip"
) -> Image:
    return Image.fromarray(channel_to_image_numpy(channel, colormap=colormap, normalization=normalization))


def channel_to_image_numpy(
        channel: torch.Tensor,
        *,
        colormap: int = cv2.COLORMAP_SUMMER,
        normalization: Literal["arctan", "linear", "clip"] | Callable[[np.ndarray], np.ndarray] = "clip"
) -> np.ndarray:
    """
    :param channel: [W x H]
    :param colormap: OpenCV-Colormap used for coloring
    :param normalization: "arctan" | "linear" | "clip" or function that maps to  a ndarray with values between [0..1]
    :return: ndarray[W x H x 3]
    """
    img: np.ndarray = channel.detach().cpu().numpy()  # img -> ndarray[W x H]
    if normalization == "arctan":  # Arctan-norm
        # Map [-inf, inf] to [-pi/2, pi/2] to [-1, 1] to [-0.5, 0.5] to [0, 1]
        img = (np.arctan(img) / (3.141592 / 2) / 2 + .5)
    elif normalization == "linear":  # Linear norm
        img = (img - np.min(img)) / (np.max(img) - np.min(img))
    elif normalization == "clip":
        img = np.clip(img, 0, 1)
    elif isinstance(normalization, callable):
        img = normalization(img)
    else:
        raise ValueError(f"Unexpected normalization. Got `{normalization}`.")

    img = (img * 255.0).astype("uint8")
    img = cv2.applyColorMap(img, colormap=colormap)
    return img


def separate_channel_numpy(
        channels: torch.Tensor,
        *,
        channel_plot_size=128,
        colormap=cv2.COLORMAP_SUMMER,
        normalization: Literal["arctan", "linear", "clip"] | Callable[[np.ndarray], np.ndarray] = "clip"
) -> np.ndarray:
    """
    Used for
        - State [C x W x H]
        - Metrics [M x W x H]
        - Target [C x W x H]
    :param channels: [N x W x H]
    :return: np.ndarray[W x H x N]
    """
    channels = clean_state(channels)
    c, w, h = channels.shape
    cviz_cols, cviz_rows = as_quadratic_shape_as_possible(c)
    cviz_w = cviz_cols * channel_plot_size
    cviz_h = cviz_rows * channel_plot_size

    channel_visualisation = np.zeros((cviz_h, cviz_w, 3), dtype=np.uint8)

    if normalization == "linear":
        # Normalize here to avoid local per channel normalization
        normalization = "clip"
        channels = (channels - torch.min(channels)) / (torch.max(channels) - torch.min(channels))

    for i in range(c):
        row = i // cviz_cols
        col = i % cviz_cols
        row_offset = row * channel_plot_size
        col_offset = col * channel_plot_size

        img = channel_to_image_numpy(channels[i], colormap=colormap, normalization=normalization)
        img = cv2.resize(img, (channel_plot_size, channel_plot_size))
        channel_visualisation[
        row_offset: row_offset + channel_plot_size,
        col_offset: col_offset + channel_plot_size
        ] = img
    return channel_visualisation


def separate_channel_pillow(
        channels: torch.Tensor,
        *,
        colormap=cv2.COLORMAP_SUMMER,
        normalization: Literal["arctan", "linear", "clip"] | Callable[[np.ndarray], np.ndarray] = "clip"
) -> list[Image]:
    """
    Used for
        - State [C x W x H]
        - Metrics [M x W x H]
        - Target [C x W x H]
    :param channels: [N x W x H]
    :return: np.ndarray[W x H x N]
    """
    c, w, h = channels.shape
    images = []
    for i in range(c):
        img = channel_to_image_numpy(channels[i], colormap=colormap, normalization=normalization)
        images.append(Image.fromarray(img))
    return images


def separate_channel_plot(
        channels: torch.Tensor,
        names: Optional[Iterable[str]] = None,
        *,
        dpi=200,
        min_val: Optional[float] = None,
        max_val: Optional[float] = None,
        colormap: Optional[mcolors.Colormap | str] = None
) -> plt.Figure:
    """
    Creates a new figure with specified dpi and plots all channels for first batch_no with a given colormap.
    Used for
        - State [(1 x) C x W x H]
        - Metrics [(1 x) M x W x H]
        - Target [(1 x) C x W x H]
    :param channels: [N x W x H]
    :param names: N strings used as caption
    :param dpi: dpi used
    :param colormap:
    :return: plt.Figure
    """
    # Parameter validation
    channels = clean_state(channels)
    if names is None:
        names = [None] * len(channels)
    if len(names) != len(channels):
        raise ValueError(f"Length of channels does not match length of names ({len(channels)} != {len(names)})")

    # Create figure
    n_channels = len(channels)
    w, h = as_quadratic_shape_as_possible(n_channels)

    # Plotting
    if min_val is None:
        min_val = channels.min()
    if max_val is None:
        max_val = channels.max()

    fig, axs = plt.subplots(h, w, squeeze=False, figsize=(w + 1, h + 1), dpi=dpi)

    # Remove ticks
    for ax1 in axs:
        for ax in ax1:
            ax.set_xticks([])
            ax.set_yticks([])
            ax.axis("off")

    # Plot images
    for i in range(len(channels)):
        ax = axs[i // w, i % w]
        name = names[i]
        if i < len(channels):
            channel = channels[i]
            if name is None:
                ax.set_title(f"Channel {i}")
            else:
                ax.set_title(name)
            ax.imshow(channel, vmin=min_val, vmax=max_val, cmap=colormap)

    fig.tight_layout(pad=0)
    return fig


def to_bgr_image(image: np.ndarray):
    """
    Processes a Gray/BGR/BGRA image to BGR
    :param image: np.ndarray[B x W x H x 1/3/4]
    :return: np.ndarray[B x W x H x 3]
    """
    B, W, H, C = image.shape
    if C == 3:
        return image
    elif C == 4:
        return blend_alpha(image)
    elif C == 1:
        return image.repeat(3, axis=3)

    raise TypeError(f"Unknown type conversion from {C} channels to BGR (image.shape={image.shape})")


def generate_video_array(
        nca: NCA,
        state: torch.Tensor,
        *,
        visualizer: StateVisualizer = RGBVisualizer(),
        steps: int = 300,
) -> tuple[dict[str, np.ndarray], torch.Tensor]:
    """
    :param nca: callable that converts `state` to `next_state`
    :param state: tensor [B, C, W, H]
    :param steps: N steps
    :param visualizer: StateVisualizer that transform a state into images
    :return: dict[name, np.ndarray[N, W, H, 3]] (BGR), final_state Tensor: [B, C, W, H]
    """
    device = nca.device

    first_frame = visualizer(state)
    results = {name: [image] for name, image in first_frame.items()}

    # Simulate n steps
    with torch.no_grad():
        state = state.to(device)
        B, _, width, height = state.shape
        for i in range(steps):
            t_float = 1 - (i / steps)
            state = nca(state, t=torch.as_tensor([t_float] * B))
            visualization = visualizer(state)
            for name, image in visualization.items():
                results[name].append(image)

    for name, images in results.items():
        results[name] = np.stack(images, axis=0)

    return results, state


def run_inference(
        nca: NCA,
        state: torch.Tensor,
        *,
        steps: int = 300,
) -> tuple[dict[str, np.ndarray], torch.Tensor]:
    """
    :param nca: callable that converts `state` to `next_state`
    :param state: tensor [B, C, W, H]
    :param steps: N steps
    :return: final_state Tensor: [B, C, W, H]
    """
    device = nca.device

    # Simulate n steps
    with torch.no_grad():
        state = state.to(device)
        B, _, width, height = state.shape
        for i in range(steps):
            t_float = 1 - (i / steps)
            state = nca(state, t=torch.as_tensor([t_float] * B))

    return state


def video_timeline_preparation(batched_image_timeline: np.ndarray, padding=5) -> np.ndarray:
    """
    Applys to_bgr_image and np_make_grid to a batch of images
    :param batched_image_timeline: np.ndarray[T x B x W x H x 3]
    :return: np.ndarray[T x 3 x W x H]
    """
    T, B, W, H, C = batched_image_timeline.shape
    nrows = int(B ** 0.5 + 0.49)
    first_frame = np_make_grid(batched_image_timeline[0], nrow=nrows, padding=padding, pad_value=255)
    W_new, H_new, C_ = first_frame.shape
    result = np.zeros((T, W_new, H_new, C_), dtype=batched_image_timeline.dtype)
    result[0] = first_frame
    for i, frame in enumerate(batched_image_timeline[1:]):
        result[i + 1] = np_make_grid(frame, nrow=nrows, padding=padding, pad_value=255)

    bgr_videos = to_bgr_image(result)
    return bgr_videos.transpose((0, 3, 1, 2))  # Apparently wandb wants to have [T x C x W x H] for videos...


def _legacy_generate_video_array(
        nca: NCA,
        state: torch.Tensor,
        *,
        steps: int = 300,
        rgb_channels: slice = slice(0, 3)
) -> np.ndarray:
    """
    :param nca: callable that converts `state` to `next_state`
    :param state: tensor [1, C, W, H]
    :param steps: N steps
    :param rgb_channels: slice of channels which should be interpreted as rgb
    :return: [N, 3, W, H] (RGB)
    """
    device = nca.device
    warnings.warn("You are using an old version of generate_video_array!")

    # Simulate n steps
    with torch.no_grad():
        state = state.to(device)
        B, _, width, height = state.shape
        result = torch.zeros((steps + 1, length_of_slice(rgb_channels), width, height), device=device)
        result[0] = state[0, rgb_channels]
        for i in range(steps):
            t_float = 1 - (i / steps)
            state = nca(state, t=torch.as_tensor([t_float] * B))
            result[i + 1] = state[0, rgb_channels]

    # Post-processing:
    # 1. Clamp between 0 and 1
    # 2. Scale between 0 and 255
    # 3. Convert to numpy uint8 array
    result = (torch.clamp(result, 0, 1) * 255).cpu().numpy().astype("uint8")
    return result


def blend_alpha(images: torch.Tensor | np.ndarray):
    """
    Source: ChatGPT
    :param images: Tensor[(T) x B x 3/4 x W x H] | np.ndarray[(T) x B x W x H x 3/4]
    :return: Tensor[(T) x B x 3 x W x H] | np.ndarray[(T) x B x W x H x 3]
    """
    if isinstance(images, torch.Tensor):
        if len(images.shape) == 5:
            rgb = images[:, :, :3]
            alpha = images[:, :, 3:4]
        elif len(images.shape) == 4:
            rgb = images[:, :3]
            alpha = images[:, 3:4]
        else:
            raise ValueError(f"Invalid shape for tensor: {images.shape}")
    elif isinstance(images, np.ndarray):
        if len(images.shape) == 5:
            rgb = images[:, :, :, :, :3]
            alpha = images[:, :, :, :, 3:4]
        elif len(images.shape) == 4:
            rgb = images[:, :, :, :3]
            alpha = images[:, :, :, 3:4]
        else:
            raise ValueError(f"Invalid shape for ndarray: {images.shape}")
    else:
        raise TypeError("Expects either tensor or ndarray")

    # No alpha channel. E.g. Shape = [(T) x B x 0 x W x H]
    if np.prod(alpha.shape) == 0:
        return images

    # Normalize alpha channel values between 0 and 1
    alpha = alpha / 255.0

    # Apply alpha blending
    blended_image = rgb * alpha + (1 - alpha) * 255.0
    if isinstance(blended_image, np.ndarray):
        return blended_image.astype("uint8")
    return blended_image


import numpy as np


def np_make_grid(
        image,
        nrow=8,
        padding=2,
        normalize=False,
        value_range=None,
        scale_each=False,
        pad_value=0.0,
):
    """
    Make a grid of images.

    Args:
        image (ndarray or list): 4D mini-batch ndarray of shape (B x H x W x C)
            or a list of images all the same size.
        nrow (int, optional): Number of images displayed in each row of the grid.
            The final grid size is ``(B / nrow, nrow)``. Default: ``8``.
        padding (int, optional): amount of padding. Default: ``2``.
        normalize (bool, optional): If True, shift the image to the range (0, 1),
            by the min and max values specified by ``value_range``. Default: ``False``.
        value_range (tuple, optional): tuple (min, max) where min and max are numbers,
            then these numbers are used to normalize the image. By default, min and max
            are computed from the tensor.
        scale_each (bool, optional): If ``True``, scale each image in the batch of
            images separately rather than the (min, max) over all images. Default: ``False``.
        pad_value (float, optional): Value for the padded pixels. Default: ``0``.

    Returns:
        grid (ndarray): the ndarray containing grid of images.
    """
    # if list of tensors, convert to a 4D mini-batch ndarray
    if isinstance(image, list):
        image = np.stack(image, axis=0)

    if image.ndim == 2:  # single channel, single image H x W -> H x W x C
        image = np.expand_dims(image, axis=2)
    if image.ndim == 3:  # single image H x W x C -> B x H x W x C
        image = np.expand_dims(image, axis=0)

    if image.ndim == 4 and image.shape[3] == 1:  # single-channel images
        image = np.concatenate((image, image, image), axis=3)

    if normalize is True:
        image = image.copy()  # avoid modifying tensor in-place
        if value_range is not None and not isinstance(value_range, tuple):
            raise TypeError("value_range has to be a tuple (min, max) if specified. min and max are numbers")

        def norm_ip(img, low, high):
            img = np.clip(img, low, high)
            img = (img - low) / max(high - low, 1e-5)
            return img

        def norm_range(t, value_range):
            if value_range is not None:
                return norm_ip(t, value_range[0], value_range[1])
            else:
                return norm_ip(t, float(np.min(t)), float(np.max(t)))

        if scale_each is True:
            for i in range(len(image)):  # loop over mini-batch dimension
                image[i] = norm_range(image[i], value_range)
        else:
            image = norm_range(image, value_range)

    if not isinstance(image, np.ndarray):
        raise TypeError("image should be of type np.ndarray")
    if image.shape[0] == 1:
        return image.squeeze(0)

    # make the mini-batch of images into a grid
    nmaps = image.shape[0]
    xmaps = min(nrow, nmaps)
    ymaps = int(np.ceil(float(nmaps) / xmaps))
    height, width = int(image.shape[1] + padding), int(image.shape[2] + padding)
    num_channels = image.shape[3]
    grid = np.full((height * ymaps + padding, width * xmaps + padding, num_channels), pad_value, dtype=image.dtype)
    k = 0
    for y in range(ymaps):
        for x in range(xmaps):
            if k >= nmaps:
                break
            grid[y * height + padding: (y + 1) * height, x * width + padding: (x + 1) * width] = image[k]
            k += 1
    return grid



import torch

from ncalib.trainer import NCA_StoppingCriterion
from ncalib.trainer.dataset import DataSample
from ncalib.trainer.loss_function import LossResult


class MaxEpochs(NCA_StoppingCriterion):
    def __init__(self, max_epochs: int = 1000):
        super().__init__(f"Reached max epochs ({max_epochs})")
        self.max_epochs = max_epochs

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        return trainer.epoch >= self.max_epochs

    def __len__(self):
        return self.max_epochs


class MaxTime(NCA_StoppingCriterion):
    def __init__(self, time_limit: float = 600):
        super().__init__(f"Reached time_limit ({time_limit} seconds)")
        self.time_limit = time_limit  # in seconds

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        if trainer._time_started is None:
            return False  # Training not started yet

        return trainer.training_duration >= self.time_limit

    def __len__(self):
        return self.time_limit


class LossIsNaN(NCA_StoppingCriterion):
    def __init__(self):
        super().__init__("Loss is not a number (NaN)", is_fatal=True)

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        if loss is None:
            # Training not started yet
            return False
        return bool(not torch.isfinite(loss.total_loss(ignore_inf=True)))


class LossSmallerThan(NCA_StoppingCriterion):
    def __init__(self, threshold: float):
        super().__init__(f"Loss below threshold ({threshold})", is_fatal=False)
        self.threshold = threshold

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        if loss is None:
            # Training not started yet
            return False
        if not torch.isfinite(loss.total_loss(ignore_inf=True)):
            return False
        return loss.total_loss() < self.threshold



import logging
from typing import Literal

import torch
from omegaconf import DictConfig

from ncalib import ChainedPerception, NCA_Perception, NCA
from ncalib.nca.models.basic import BasicNCAModel
from ncalib.nca.modular_nca import ModularNCA
from ncalib.nca.perception.filter import IdentityPerception, SobelXPerception, SobelYPerception, LaPlace1Perception, \
    LaPlace2Perception
from ncalib.utils.defaults import DEFAULT_DEVICE

logger = logging.getLogger("utils.nca")


def build_perception(
        *,
        channels=16,
        identity=True,
        sobel_x=True,
        sobel_y=True,
        laplace1=False,
        laplace2=False,
        padding_mode: Literal["constant", "reflect", "replicate", "circular"] = "replicate"
):
    logger.debug(f"Perception: Build chained perception for {channels} channels")
    perception = ChainedPerception(channels)
    if identity:
        logger.debug(f"Perception: Add identity")
        perception += IdentityPerception(channels)
    if sobel_x:
        logger.debug(f"Perception: Add Sobel X")
        perception += SobelXPerception(channels, padding_mode=padding_mode)
    if sobel_y:
        logger.debug(f"Perception: Add Sobel Y")
        perception += SobelYPerception(channels, padding_mode=padding_mode)
    if laplace1:
        logger.debug(f"Perception: Add LaPlace (v1)")
        perception += LaPlace1Perception(channels, padding_mode=padding_mode)
    if laplace2:
        logger.debug(f"Perception: Add LaPlace (v2)")
        perception += LaPlace2Perception(channels, padding_mode=padding_mode)

    return perception


def build_modular_nca(
        channels: int,
        perception: NCA_Perception,
        model: DictConfig,
        update_probability: float = 0.5,
        device: torch.device | str = DEFAULT_DEVICE,

) -> NCA:
    logger.debug(f"NCA: Build NCA with {channels} channels and {update_probability=} for device {device}")
    nca_model = BasicNCAModel(
        perception.output_size_for_n_channels(channels),
        channels,
        hidden_channels=model["hidden_channels"]
    )
    nca = ModularNCA(
        channels,
        perception,
        nca_model,
        device=device
    )

    # Load state dict
    if model.state_dict:
        logger.debug(f"Load state_dict")
        nca.load_state_dict(torch.load(model.state_dict, map_location=device))
    else:
        logger.debug(f"No state_dict found for model")
    return nca



from typing import Callable, Any

import torch
from torch.utils.data import DataLoader

from ncalib.trainer.batcher import NCA_Batcher
from ncalib.trainer.batcher.pool.eviction_policies import PoolEvictionPolicy
from ncalib.trainer.batcher.pool.eviction_policies.loss_policies import LossSortEvictionPolicy
from ncalib.trainer.dataset import NCA_Dataset, DataSample
from ncalib.trainer.loss_function import LossResult
from ncalib.utils.defaults import DEFAULT_DEVICE


class PooledBatcher(NCA_Batcher):
    def __init__(
            self,
            dataset: NCA_Dataset,
            batch_size: int,
            pool_size: int,
            eviction_policy: PoolEvictionPolicy = LossSortEvictionPolicy(),
            replace_n_worst: int = 1,
            num_workers: int = 0,
            warmup: int = 0,
            collate_fn: Callable = DataSample.merge_batch,
            device: torch.device | str = DEFAULT_DEVICE,
    ):
        super().__init__(dataset, batch_size, num_workers=num_workers, collate_fn=collate_fn, device=device)
        self.dataloader = iter(
            DataLoader(
                self.dataset,
                batch_size=replace_n_worst if replace_n_worst >= 1 else 1,
                shuffle=False,  # Shuffling is done in dataset
                num_workers=self.num_workers,
                drop_last=True,
                collate_fn=self.collate_fn,
            )
        )

        self.eviction_policy = eviction_policy
        self.pool_size = pool_size
        self.replace_worst_n = replace_n_worst
        self.pool = None
        self._losses = None
        self.warmup = warmup

        self._current_replaces = 0

    def init(self):
        init_dataloader = iter(DataLoader(
            dataset=self.dataset,
            batch_size=self.pool_size,
            shuffle=False,  # Shuffling is done in dataset
            num_workers=self.num_workers,
            collate_fn=self.collate_fn
        ))
        self.pool = next(init_dataloader)
        self._losses = torch.ones(self.pool_size) * torch.inf

    def next_batch(self) -> DataSample:
        idx = torch.randint(high=self.pool_size, size=(self.batch_size,))
        sample = self.pool[idx]
        self._last_idx = idx

        # Replace worst
        self._current_replaces += self.replace_worst_n
        if self._current_replaces > 0:
            scores = self.eviction_policy(sample, self._losses[idx])
            sorted_idx = torch.argsort(scores, descending=True)
            sample[sorted_idx[:int(self._current_replaces)]] = next(self.dataloader)
            self._current_replaces -= int(self._current_replaces)

        return sample.clone().to(self.device)


    def feedback(self, data_sample: DataSample, new_state: torch.Tensor, loss: LossResult):
        if self.warmup > 0:
            self.warmup -= 1
            return
        idx = self._last_idx
        self._losses[idx] = loss.batchwise_loss().detach().to("cpu")
        self.pool[idx] = data_sample.to("cpu")

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["pool_size"] = self.pool_size
        cfg["replace_worst_n"] = self.replace_worst_n
        cfg["eviction_policy"] = self.eviction_policy.to_config_dict()
        cfg["warmup"] = self.warmup

        return cfg



import logging
from functools import cached_property
from pathlib import Path
from typing import Optional

import cv2
import numpy as np
import torch
from PIL.Image import Image
from torchvision import transforms


class RGBAImage:
    def __init__(self, rgba_target: torch.Tensor):
        """
        :param rgba_target: Tensor[4 x W x H]
        """
        if len(rgba_target.shape) != 3:
            raise ValueError("Target tensor has to have 3 dimensions")
        if rgba_target.shape[0] != 4:
            raise ValueError(f"Target tensor has to have 4 channels (rgba) (has {rgba_target.shape[0]} channels)")

        self.img = rgba_target

    # Constructors
    @classmethod
    def from_numpy(cls, image: np.ndarray) -> "RGBAImage":
        pass

    @classmethod
    def from_pillow(cls, image: np.ndarray) -> "RGBAImage":
        pass

    @classmethod
    def from_path(cls, path: Path | str) -> "RGBAImage":
        full_path = str(Path(path).absolute())
        logger = logging.getLogger(cls.__name__)
        logger.info(f"Attempting to load `{full_path}`")
        img_bgr = cv2.imread(full_path, cv2.IMREAD_UNCHANGED)

        if img_bgr is None:
            raise FileNotFoundError(full_path)

        if len(img_bgr.shape) == 2:  # Grayscale
            tensor = torch.unsqueeze(torch.from_numpy(img_bgr), dim=0)
            tensor = tensor.repeat([3, 1, 1])
        elif len(img_bgr.shape) == 3:  # RGB / RGBA
            img_rgb = np.copy(img_bgr)
            img_rgb[:, :, 0] = img_bgr[:, :, 2]
            img_rgb[:, :, 2] = img_bgr[:, :, 0]
            tensor = torch.from_numpy(img_rgb.transpose((2, 0, 1)))
        else:
            raise RuntimeError(f"Unknown format for img with shape {img_bgr.shape} -> `{img_bgr}`")

        # Add artificial alpha channel if needed
        if tensor.shape[0] == 3:
            logger.info("Adding artificial alpha channel")
            rgba_tensor = torch.ones((4, tensor.shape[1], tensor.shape[2])) * 255
            rgba_tensor[:3] = tensor
            tensor = rgba_tensor

        return RGBAImage(tensor / 255)

    # Basic properties
    @property
    def width(self):
        return self.img.shape[1]

    @property
    def height(self):
        return self.img.shape[2]

    # Image type conversions
    def as_numpy(self) -> np.ndarray:
        """
        :return: ndarray[W x H x 4] as BGRA
        """
        # TODO: Image orientation not correct
        return np.rot90(self.img.cpu().numpy().transpose((2, 1, 0))[::-1, ::-1, :3])[:, ::-1]

    def as_pillow(self) -> Image:
        return transforms.ToPILImage()(self.img)

    # == Single channels ==
    # RGBA
    @property
    def rgba_red(self) -> torch.Tensor:
        return self.img[0]

    @property
    def rgba_green(self) -> torch.Tensor:
        return self.img[1]

    @property
    def rgba_blue(self) -> torch.Tensor:
        return self.img[2]

    @property
    def rgba_alpha(self) -> torch.Tensor:
        return self.img[3]

    @cached_property
    def rgb_max(self) -> torch.Tensor:
        return torch.max(self.img[:3], dim=0).values

    @cached_property
    def rgb_min(self) -> torch.Tensor:
        return torch.min(self.img[:3], dim=0).values

    @cached_property
    def rgb_delta(self) -> torch.Tensor:
        return self.rgb_max - self.rgb_min

    # Modification
    def pad(
            self,
            padding: int,
    ) -> "RGBAImage":
        new_img = torch.zeros((4, self.width + padding * 2, self.height + padding * 2))
        new_img[:, padding:-padding, padding:-padding] = self.img

        return RGBAImage(new_img)

    def resize(
            self,
            width: int,
            height: Optional[int] = None,
    ) -> "RGBAImage":
        if height is None:
            aspect = self.width / self.height
            height = int(width / aspect)

        if width == self.width and height == self.height:
            new_img = self.img
        else:
            new_img = transforms.ToTensor()(self.as_pillow().resize((width, height)))

        return RGBAImage(new_img)

    def scale(
            self,
            scale: float,
    ) -> "RGBAImage":
        pil_image = self.as_pillow()
        new_width = int(pil_image.width * scale)
        new_height = int(pil_image.height * scale)

        return self.resize(new_width, new_height)


def gt_anomaly_percentage(ground_truth: torch.Tensor) -> float:
    return (torch.sum(ground_truth) / (ground_truth.shape[0] * ground_truth.shape[1])).item()


def target_id_from_path(path: str | Path) -> str:
    return Path(path).stem



import abc
from pathlib import Path
from typing import Optional

import torch
import wandb
from wandb.apis.public import Run
from wandb.sdk.lib import RunDisabled

from ncalib.trainer import NCA_StoppingCriterion, NCA_Trainer
from ncalib.trainer.dataset import DataSample
from ncalib.trainer.logger import NCA_Logger
from ncalib.trainer.loss_function import LossResult


class WandBLogger(NCA_Logger, abc.ABC):
    """
    Extends the normal logger so that it basically gets disabled when wandb is disabled by replacing self.log with noop
    """

    def __init__(
            self,
            *,
            epoch_interval: int = 1000,
            log_final: bool = False
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final)
        self.run: Optional[Run] = None

    def init_training(self, trainer: NCA_Trainer, stopping_criterion: NCA_StoppingCriterion):
        super().init_training(trainer, stopping_criterion)
        if wandb.run is None:
            raise RuntimeError("WandB not initialized yet. Please call .wandb.init() when using WandBLogger!")
        self.run = wandb.run
        if isinstance(self.run, RunDisabled):
            self.logger.warn(f"WandB is disabled. Disabling {self}")
            self.log = self.log_noop

    def log_noop(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        """ Replaces self.log if mode =="disabled" """
        pass


class WandBModelLogger(WandBLogger):
    def __init__(
            self,
            *,
            leading_zeros: int = 5,
            epoch_interval: int = 1000,
            log_final: bool = False
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final)
        self.leading_zeros = leading_zeros
        self.latest_model_path = None

    def init_training(self, trainer: NCA_Trainer, stopping_criterion: NCA_StoppingCriterion):
        super().init_training(trainer, stopping_criterion)
        wandb.watch(self.trainer.nca, log_freq=self.epoch_interval)

    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        log_path = Path(self.run.dir) / "models" / f"model-{self.trainer.epoch:0{self.leading_zeros}d}.nca"
        log_path.parent.mkdir(exist_ok=True, parents=True)
        torch.save(self.trainer.nca.as_full_config_dict(), log_path)
        wandb.save(str(log_path.absolute()), base_path=self.run.dir)
        self.latest_model_path = log_path



from typing import Optional

import torch

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.utils import slice_to_config, to_binary_list


class GenomeSeedFactory(NCA_SeedFactory):
    def __init__(
            self,
            genome_channels: slice,
            width: int = 64,
            channels: int = 16,
            height: Optional[int] = None,
            visible_channels: slice = slice(0, 3),
            deterministic=False
    ):
        super().__init__(width=width, height=height, channels=channels)
        self.visible_channels = visible_channels
        self.genome_channels = genome_channels
        self.deterministic = deterministic

    def __call__(
            self,
            n: int = 1,
            *,
            device=DEFAULT_DEVICE,
            genome: Optional[torch.Tensor | list[float]] = None
    ) -> torch.Tensor:
        seed = torch.zeros([n, self.channels, self.width, self.height], dtype=torch.float32, device=device)
        seed[:, self.visible_channels, self.width // 2, self.height // 2] = 1.0
        genome_place = seed[:, self.genome_channels, self.width // 2, self.height // 2]
        if genome is None:
            if self.deterministic:
                genome = torch.zeros_like(genome_place)
                for i in range(len(genome)):
                    genome[i] = torch.as_tensor(to_binary_list(i, min_length=len(genome[i])))
            else:
                genome = torch.randint_like(genome_place, 0, 2)

        else:
            # Check shape
            B, G = genome_place.shape
            genome = torch.as_tensor(genome, dtype=torch.float32)
            if len(genome.shape) == 1:
                # Expects to have genome that is same for all batches
                if not genome.shape[0] == G:
                    raise ValueError(f"Invalid genome size. Expects genome of size {G}, but got shape {genome.shape}")

                genome = genome[None]
            elif len(genome.shape) == 2:
                if not genome.shape == genome_place.shape:
                    raise ValueError(
                        f"Invalid genome size. "
                        f"Expects genome of size {genome_place.shape}, but got shape {genome.shape}"
                    )
        seed[:, self.genome_channels, self.width // 2, self.height // 2] = torch.as_tensor(genome, dtype=torch.float32)
        return seed

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["genome_channels"] = slice_to_config(self.genome_channels)
        cfg["visible_channels"] = slice_to_config(self.visible_channels)
        return cfg



import abc
from typing import Any

import torch
import torch.nn as nn

from ncalib.utils.utils import NCA_Base


class NCA_Model(NCA_Base, nn.Module, abc.ABC):
    def __init__(self, input_size: int, output_size: int):
        super().__init__()
        nn.Module.__init__(self)

        self.input_size = input_size
        self.output_size = output_size

    @abc.abstractmethod
    def forward(self, perception: list[torch.Tensor], **kwargs) -> torch.Tensor:
        pass

    def reset(self):
        raise NotImplementedError(f"reset missing in {self.__class__.__name__}")

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["input_size"] = self.input_size
        cfg["output_size"] = self.output_size
        return cfg

    def required_kwargs(self) -> set[str]:
        return set()
